""" Websocket-part of the Werk24 client
"""
import json
from types import TracebackType
from typing import Optional, Type, AsyncGenerator

import websockets
from pydantic import ValidationError
from websockets.client import WebSocketClientProtocol
from werk24.exceptions import ServerException, UnauthorizedException
from werk24.models.techread import W24TechreadCommand, W24TechreadMessage
from werk24.auth_client import AuthClient
import logging

# make the logger
logger = logging.getLogger("w24_techread_client")

class TechreadClientWss:
    """
    TechreadClient subpart that handles the websocket
    communication with the server.
    """

    def __init__(
        self, 
        techread_server_wss: str, 
        techread_version: str,
        wss_close_timeout: Optional[float] = 600
    ):
        """
        Initialize the TechreadClientWss

        Args:
        ----
        - techread_server_wss (str): The server to connect to
        - techread_version (str): The version of the server
        - wss_close_timeout (Optional[float], optional): The time to wait for the
            websocket to close. Defaults to 600.
        """
        logger.debug("Using Websockets version: %s", websockets.__version__)
        logger.debug(
            "Creating TechreadClientWss with server %s and version %s. Websocket timeout: %d",
            techread_server_wss,
            techread_version,
            wss_close_timeout
        )

        self._techread_server_wss = techread_server_wss
        self._techread_version = techread_version
        self._techread_session_wss: Optional[WebSocketClientProtocol] = None
        self.endpoint = f"wss://{self._techread_server_wss}/{self._techread_version}"
        self._auth_client = None
        self.wss_close_timeout = wss_close_timeout

    async def __aenter__(self) -> "TechreadClientWss":
        """
        Enter the session with the wss server

        Raises:
        ------
        - RuntimeError: Raise when the developer enters the session
            without having specified a token

        Returns:
        -------
        - TechreadClientWss: instance with activated session
        """
        logger.debug(f"Entered the session with the server {self._techread_server_wss}")
        headers = self._auth_client.get_auth_headers()
        self._techread_session_wss = await websockets.connect(
            self.endpoint,
            extra_headers=headers,
            close_timeout=self.wss_close_timeout,
        )
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Close the session
        """
        logger.debug(f"Exiting the session with the server {self._techread_server_wss}")
        if self._techread_session_wss is not None:
            await self._techread_session_wss.close()

    def register_auth_client(self, auth_client: AuthClient) -> None:
        """
        Register the reference to the authentication service

        Args:
        ----
        - auth_client (AuthClient): Reference to Authentication
            client
        """
        self._auth_client = auth_client

    async def send_command(self, action: str, message: str = "{}") -> None:
        """                 
        Sends a command to the websocket.

        This method wraps the given action and message into a 
        W24TechreadCommand object, serializes it to JSON, and sends it to the 
        server via the websocket.

        Args:
        ----
        - action (str): The action requested by the client.
        - message (str, optional): Additional data to send along with the action. 
            Defaults to "{}". It should be a JSON-encoded string for easy 
            expansion.

        Raises:
        ------
        - RuntimeError: Raised if the method is called before initializing the 
            profile (i.e., if the websocket session is not established).
        """
        logger.debug(f"Sending command with action {action}")

        # Ensure the websocket session is active
        if not self._techread_session_wss:
            raise RuntimeError(
                "Profile entry is required before sending commands. "
                "Please call the appropriate method to enter the profile."
            )

        # Create the command object
        command = W24TechreadCommand(action=action, message=message)
        logger.debug(f"Sending command: %s", command.model_dump_json())

        # Send the serialized command to the websocket server
        await self._techread_session_wss.send(command.model_dump_json())


    async def recv_message(self) -> W24TechreadMessage:
        """
        Receive a message from the websocket and interpret the result as W24TechreadMessage

        Raises:
        ------
        - RuntimeError: When trying to send a command without having entered the profile.

        Returns:
        -------
        - W24TechreadMessage: interpreted message
        """

        # make sure that we have an AuthClient
        if self._techread_session_wss is None:
            raise RuntimeError(
                "You need to call enter the profile before receiving command"
            )

        # wait for the websocket to say something and interpret the message
        message_raw = str(await self._techread_session_wss.recv())
        logger.debug(f"Received message: {message_raw}")
        message = await self._process_message(message_raw)
        return message

    @staticmethod
    async def _process_message(message_raw: str) -> W24TechreadMessage:
        """
        Interpret the raw websocket message and
        turn it into a W24TechreadMessage

        Args:
        ----
        - message_raw (str): Raw message

        Raises:
        ------
        - UnauthorizedException: Exception is raised
            when you requested an action that you
            have no privileges for (or that does
            not exist)

        - ServerException: Exception is raised when
            the server did not respond as expected

        Returns:
        -------
        - W24TeachreadMessage: interpreted message
        """
        logger.debug(f"Processing message: {message_raw}")
        try:
            return W24TechreadMessage.model_validate_json(message_raw)

        except ValidationError as exception:
            # The Gateway responds with the format
            # {"message": str, "connectionId":str, "requestId":str}
            # Obtain the message
            response = json.loads(message_raw)
            error_message = response.get("message")

            # raise a specific exception if the
            # requested action was forbidden
            if error_message == "Forbidden":
                raise UnauthorizedException("Requested Action forbidden") from exception

            # otherwise fail with an UnknownException
            raise ServerException(
                f"Unexpected server response '{message_raw}'."
            ) from exception

    async def listen(self) -> AsyncGenerator[W24TechreadMessage, None]:
        """
        Simple generator that waits for messages on the websocket, interprets
        them and yields them

        Yields:
        ------
        - W24TechreadMessage: interpreted message from the socket

        Raises:
        ------
        - RuntimeError: Raise when the developer tries to send a command
                without entering the profile
        """
        logger.debug("Listening for messages")

        # make sure that we have an AuthClient
        if self._techread_session_wss is None:
            raise RuntimeError("You need to call enter the profile before listening")
    
        # wait for incoming messages
        stop = False
        try:
            while not stop:
                message_raw = str(await self._techread_session_wss.recv())
                message = await self._process_message(message_raw)
                yield message
        except (websockets.exceptions.ConnectionClosedError, websockets.exceptions.ConnectionClosedOK) as exception:
            return



from .position import W24PositionedFeature
from pydantic import BaseModel


class W24Balloon(BaseModel):
    """Balloon used for marking a feature on a drawing.

    Attributes:
    ----------
    placement (W24PlacementEllipse): Placement of the Balloon on the different thumbnails
    label (str): Label of the Balloon. Typically an ascending number.
    """

    position: W24PositionedFeature
    label: str



""" HTTPS-part of the Werk24 client
"""

import io
import uuid
from werk24.exceptions import SSLCertificateError
import json
import urllib.parse
from pydantic import UUID4
from werk24.models.ask import W24AskUnion
from types import TracebackType
from typing import Dict, Optional, Type, Union, List
from io import BufferedReader
import aiohttp
from pydantic import HttpUrl
from werk24.auth_client import AuthClient
from werk24.crypt import encrypt_with_public_key, decrypt_with_private_key
from werk24.exceptions import (
    BadRequestException,
    RequestTooLargeException,
    ResourceNotFoundException,
    ServerException,
    UnauthorizedException,
    UnsupportedMediaType,
    InsufficientCreditsException,
)
import logging
import ssl
import certifi
from werk24.models.techread import W24TechreadWithCallbackPayload
from werk24.models.helpdesk import W24HelpdeskTask
from werk24.models.techread import W24PresignedPost
from werk24._version import __version__

EXCEPTION_CLASSES = {
    range(200, 300): None,
    range(400, 401): BadRequestException,
    range(401, 404): UnauthorizedException,
    range(404, 405): ResourceNotFoundException,
    range(413, 414): RequestTooLargeException,
    range(415, 416): UnsupportedMediaType,
    range(429, 430): InsufficientCreditsException,
    range(300, 400): ServerException,
    range(500, 600): ServerException,
    range(416, 500): ServerException,
}

# make the logger
logger = logging.getLogger("w24_techread_client")

class TechreadClientHttps:
    """Translation map from the server response
    to the W24TechreadArchitectureStatus enum
    """

    def __init__(
        self,
        techread_version: str,
        support_base_url: str,
        local_public_key: Optional[bytes] = None,
    ):
        """
        Initialize a new session with the https server.

        Args:
        ----
        techread_server_https (str): Domain of the Techread https server
        techread_version (str): Techread Version
        support_base_url (str): Base URL for support requests
        local_public_key (Optional[bytes]): Local public key that allows the
            server to encrypt the result of the techread request.
        """
        logger.debug(f"Creating TechreadClientHttps with version {techread_version}")
        self._techread_version = techread_version
        self._auth_client: Optional[AuthClient] = None
        self.support_base_url = support_base_url
        self.local_public_key = local_public_key

    def _make_session(self, timeout_seconds=30) -> aiohttp.TCPConnector:
        """
        Make the connector for the session.

        Returns:
        --------
        - aiohttp.TCPConnector: Connector for the session
        """
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = aiohttp.TCPConnector(ssl=ssl_context)
        timeout = aiohttp.ClientTimeout(
            total=None,
            sock_connect=timeout_seconds,
            sock_read=timeout_seconds,
        )
        return aiohttp.ClientSession(
            timeout=timeout,
            connector=connector,
        )

    async def __aenter__(self) -> "TechreadClientHttps":
        """
        Create a new HTTP session that is being used for the whole connection.
        Be sure to keep the session alive.

        Raises:
        ------
        - RuntimeError: Exception raised when the developer tries to start the
            session without a token.

        Returns:
        -------
        - TechreadClientHttps: Instance of the class itself with active session.
        """
        logger.debug("Entered the session with the server %s", self.support_base_url)
        if self._auth_client is None:
            raise RuntimeError("No AuthClient was registered")

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Close the session.

        Args:
        ----
        - exc_type (Optional[Type[BaseException]]): The type of exception that
            caused the context manager to be exited.
        - exc_value (Optional[BaseException]): The instance of the exception that
            caused the exit.
        - traceback (Optional[TracebackType]): A traceback from the exception.

        Returns:
        --------
        None
        """
        pass

    def register_auth_client(self, auth_client: AuthClient) -> None:
        """Register the reference to the authentication service

        Args:
        ----
        - auth_client (AuthClient): Reference to Authenticatio client
        """
        self._auth_client = auth_client

    async def upload_associated_file(
        self,
        presigned_post: W24PresignedPost,
        content: Union[bytes, io.BufferedReader, None],
        public_server_key: Optional[bytes] = None,
    ) -> None:
        """
        Uploads an associated file to the API.

        This can either be a technical drawing or a 3D model.

        NOTE:
        ----- 
        The complete message size must not be larger than 10 MB.

        Args:
        ----
        - presigned_post (W24PresignedPost): Presigned post object for file upload.
        - content (Optional[bytes]): Content of the file as bytes.
        - public_server_key (Optional[bytes], optional): Public key of the server.

        Raises:
        -------
        - Various exceptions based on the issues with API, authentication
            or the requested file.
        """
        logger.debug("Uploading associated file to the server")

        # ignore if payload is empty
        if content is None:
            logger.debug("No content to upload")
            return

        # encrypt the content if we have the public key of the server
        if public_server_key is not None:
            logger.debug("Encrypting the content with the public key of the server")
            content = encrypt_with_public_key(public_server_key, content)

        # generate the form data by merging the presigned
        # fields with the file
        form = aiohttp.FormData()
        for key, value in presigned_post.fields_.items():
            form.add_field(key, value)
        form.add_field("file", content)

        # create a new fresh session that does not
        # carry the authentication token
        logger.debug("Uploading the file to the server with the presigned post")
        presigned_post_str = str(presigned_post.url)
        try:
            async with self._make_session() as session:
                async with session.post(presigned_post_str, data=form) as response:
                    self._raise_for_status(presigned_post_str, response.status)

        # Raise SSLCertificateError if the certificate is not trusted
        except aiohttp.ClientConnectorCertificateError as exception:
            raise SSLCertificateError() from exception

    async def download_payload(
        self,
        payload_url: HttpUrl,
        client_private_key_pem: Optional[bytes],
        client_private_key_passphrase: Optional[bytes] = None,
    ) -> bytes:
        """
        Return the payload from the server

        Args:
        ----
        - payload_url {HttpUrl} -- Url of the payload

        Raises:
        ------
        - RuntimeError: Hard Error that is raised when
            the function is asked to download a payload
            from an untrusted source.
            This provides some sort of protection against
            payload-injection and token-theft. When you
            see this error showing up, you should
            definitely INVESTIGATE AND LET US KNOW
            IMMEDIATELY!!!
            Call all our numbers on a Sunday morning
            at 3am if it must be. Even if its Christmas
            and Easter on the same day.
        - BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!
        - UnauthorizedException: Raised when the token
            or the requested file have expired
        - ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.
        - RequestTooLargeException: Raised when the status
            code was 413
        - UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read (because its media type
            is not supported by the API).
        - ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        - bytes: Payload
        """
        logger.debug(f"Downloading payload")

        # send the get request to the endpoint
        try:
            async with self._make_session() as session:
                response = await session.get(str(payload_url))
                self._raise_for_status(payload_url, response.status)
                raw = await response.content.read()

        # reraise the exceptions
        except (
            UnauthorizedException,  # pylint: disable=try-except-raise
            RequestTooLargeException,
            ServerException,
            BadRequestException,
            ResourceNotFoundException,
        ):
            raise
        

        if client_private_key_pem is not None:
            logger.debug("Decrypting the payload with the private key")
            return decrypt_with_private_key(
                client_private_key_pem,
                client_private_key_passphrase,
                raw,
            )
        else:
            logger.debug("Returning the raw payload. Not encrypted")

        return raw

    @staticmethod
    def _raise_for_status(url: str, status_code: int) -> None:
        """
        Raise the correct exception depending on the status code.

        Args:
        ----
        - url (str): The requested URL
        - status_code (int): The received response status code

        Raises:
        ------
        - BadRequestException: Raised when the request body cannot be interpreted. 
            This normally indicates that the API version has been updated and that
            we missed a corner case. If you encounter this exception, it is very 
            likely our mistake. Please get in touch!
        - UnauthorizedException: Raised when the token or the requested file have expired
        - ResourceNotFoundException: Raised when you are requesting an endpoint that does 
            not exist. Again, you should not encounter this, but if you do, let us know.
        - RequestTooLargeException: Raised when the status code was 413
        - UnsupportedMediaTypException: Raised when the file you submitted cannot be read
            (because its media type is not supported by the API).
        - ServerException: Raised for all other status codes that are not 2xx
        - InsufficentCreditsException: Raised when the user does not have enough credits
        """
        for key, exception_class in EXCEPTION_CLASSES.items():
            logger.debug("Request to '%s' returned status code %s", url, status_code)
            if status_code in key:
                if not (200 <= status_code < 300):
                    logger.warning("Request failed with status code %s", status_code)
                if exception_class is not None:
                    raise exception_class(
                        f"Request failed '{url}' with code {status_code}"
                    )
                return None

        # If the response code is anything other than unauthorized or 200 (OK), we trigger a ServerException.
        raise ServerException(f"Request failed '{url}' with code {status_code}")

    async def create_helpdesk_task(self, task: W24HelpdeskTask) -> W24HelpdeskTask:
        """
        Create a Helpdesk ticket.

        Args:
        ----
        task (W24HelpdeskTask): Helpdesk task to be created

        Raises:
        ------
        - BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!
        - UnauthorizedException: Raised when the token
            or the requested file have expired
        - ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.
        - RequestTooLargeException: Raised when the status
            code was 413
        - UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read(because its media type
            is not supported by the API).
        - ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        - W24HelpdeskTask: Created helpdesk task with an updated task_id.
        """
        logger.debug("Creating a helpdesk task")

        headers = self._make_helpdesk_headers()
        url = self._make_support_url("helpdesk/create-task")
        async with self._make_session() as session:
            response = await session.post(url, json=task.json(), headers=headers)
            self._raise_for_status(url, response.status)

        # return the updated task
        return W24HelpdeskTask.parse_raw(await response.text())

    def _make_support_url(self, path: str) -> str:
        """
        Make the support url for the help desk requests.

        Args:
        ----
        - path (str): Path to the endpoint

        Returns:
        -------
        - str: URL to the endpoint
        """
        return urllib.parse.urljoin(f"https://{self.support_base_url}", path)

    def _make_helpdesk_headers(self) -> Dict[str, str]:
        """
        Make the headers for the help desk requests.

        Simply the authorization header at this stage.

        Returns:
        -------
        Dict[str, str]: Help desk headers
        """
        return self._auth_client.get_auth_headers()

    async def read_drawing_with_callback(
        self,
        drawing: Union[BufferedReader, bytes],
        asks: List[W24AskUnion],
        callback_url: str,
        max_pages: int = 5,
        drawing_filename: Optional[str] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        public_key: Optional[bytes] = None,
    ) -> UUID4:
        """
        Read a drawing with a callback.

        Args:
        ----
        - drawing (Union[BufferedReader, bytes]): Drawing to be read
        - asks (List[W24Ask]): List of asks
        - callback_url (str): Callback URL
        - max_pages (int, optional): Maximum number of pages to be read.
            Defaults to 5.
        - drawing_filename (Optional[str], optional): Filename of the drawing.
            Defaults to None.
        - callback_headers (Optional[Dict[str, str]], optional): Headers for the
            callback. Defaults to None.
        - public_key (Optional[bytes], optional): Public key for the client.

        Raises:
        ------
        - BadRequestException: Raised when the request body cannot be interpreted. 
            This normally indicates that the API version has been updated and that
            we missed a corner case. If you encounter this exception, it is very 
            likely our mistake. Please get in touch!
        - UnauthorizedException: Raised when the token or the requested file have 
            expired
        - ResourceNotFoundException: Raised when you are requesting an endpoint that 
            does not exist. Again, you should not encounter this, but if you do, let us know.
        - RequestTooLargeException: Raised when the status code was 413
        - UnsupportedMediaTypException: Raised when the file you submitted cannot be read
            (because its media type is not supported by the API).
        - ServerException: Raised for all other status codes that are not 2xx
        - InsufficentCreditsException: Raised when the user does not have enough credits
            to perform the operation.

        Returns:
        -------
        - UUID4: Request ID
        """
        logger.debug("Reading drawing with callback")

        # Set a default drawing filename if none is provided
        drawing_filename = drawing_filename or "drawing.pdf"
        logger.debug("Drawing filename: %s", drawing_filename)

        # validate the payload locally. This is not strictly necessary
        # but it is a good way to catch errors early.
        payload = W24TechreadWithCallbackPayload(
            asks=asks,
            callback_url=callback_url,
            callback_headers=callback_headers,
            max_pages=max_pages,
            client_version=__version__,
            drawing_filename=drawing_filename,
            public_key=public_key,
        )

        # create the form data
        data = aiohttp.FormData()
        data.add_field("drawing", drawing, filename=drawing_filename)
        for key, value in payload.model_dump(mode="json").items():
            data.add_field(key, json.dumps(value))

        # send the request
        headers = self._auth_client.get_auth_headers()
        url = self._make_support_url("techread/read-with-callback")
        async with self._make_session() as session:
            response = await session.post(url, data=data, headers=headers)
            self._raise_for_status(url, response.status)
            response_json = await response.json(content_type=None)

        try:
            return uuid.UUID(response_json["request_id"])
        except (ValueError, KeyError):
            raise BadRequestException(f"Request failed: {response_json}")




from typing import Literal

from werk24.models.property.base import W24Property
from werk24.models.value import W24PhysicalQuantity


class W24PropertySurfaceArea(W24Property):
    """Surface Area

    Attributes:
        surface_area (W24PhysicalQuantityRange): Range
            of the surface area. We deliberately chose
            this to be a range - it is possible, that
            that people specify allowed ranges.
    """
    property_type: Literal["SURFACE_AREA"] = "SURFACE_AREA"
    property_subtype: Literal["SURFACE_AREA"] = "SURFACE_AREA"
    surface_area: W24PhysicalQuantity



from enum import Enum


class W24EarthLocation(str, Enum):
    """Earth Location based on the United Nations LoCodes.

    The Werk24 Earth locations are based on the list of the
    United Nations UN/LOCODE list of countries and territories.
    See: https://unece.org/trade/cefact/unlocode-code-list-country-and-territory.

    This list was last updated on 2023-04-26 and is subject to
    human history. Please be prepared your code to handle
    changes. The naming conventions follow ISO-3166-1-alpha-2:2020-12.

    NOTE: This model shall be limited to locations on Planet Earth.
    When humans start settling other planets, we might need to update this.

    NOTE: Werk24 extends this list to include other inter-country organizations
    such as NATO. To ensure that there are no naming conflicts, we add the
    prefix ORG_. We deviate from the UN/LOCODE logic and use the English
    names for all of the organizations. We do this to treat Organizations
    the same, independent of whether they opt for a latin-based or non-latin
    based script.
    """
    ORG_NATO = "North Atlantic Treaty Organization"

    AF = "Afghanistan"
    AX = "Åland Islands"
    AL = "Albania"
    DZ = "Algeria"
    AS = "American Samoa"
    AD = "Andorra"
    AO = "Angola"
    AI = "Anguilla"
    AQ = "Antarctica"
    AG = "Antigua and Barbuda"
    AR = "Argentina"
    AM = "Armenia"
    AW = "Aruba"
    AU = "Australia"
    AT = "Austria"
    AZ = "Azerbaijan"
    BS = "Bahamas"
    BH = "Bahrain"
    BD = "Bangladesh"
    BB = "Barbados"
    BY = "Belarus"
    BE = "Belgium"
    BZ = "Belize"
    BJ = "Benin"
    BM = "Bermuda"
    BT = "Bhutan"
    BO = "Bolivia"
    BQ = "Bonaire, Sint Eustatius and Saba"
    BA = "Bosnia and Herzegovina"
    BW = "Botswana"
    BR = "Brazil"
    IO = "British Indian Ocean Territory"
    BN = "Brunei Darussalam"
    BG = "Bulgaria"
    BF = "Burkina Faso"
    BI = "Burundi"
    KH = "Cambodia"
    CM = "Cameroon"
    CA = "Canada"
    CV = "Cape Verde"
    KY = "Cayman Islands"
    CF = "Central African Republic"
    TD = "Chad"
    CL = "Chile"
    CN = "China"
    CX = "Christmas Island"
    CC = "Cocos (Keeling) Islands"
    CO = "Colombia"
    KM = "Comoros"
    CG = "Congo"
    CD = "Congo, The Democratic Republic of the"
    CK = "Cook Islands"
    CR = "Costa Rica"
    CI = "Côte d'Ivoire"
    HR = "Croatia"
    CU = "Cuba"
    CW = "Curaçao"
    CY = "Cyprus"
    CZ = "Czech Republic"
    DK = "Denmark"
    DJ = "Djibouti"
    DM = "Dominica"
    DO = "Dominican Republic"
    EC = "Ecuador"
    EG = "Egypt"
    SV = "El Salvador"
    GQ = "Equatorial Guinea"
    ER = "Eritrea"
    EE = "Estonia"
    SZ = "Eswatini"
    ET = "Ethiopia"
    FK = "Falkland Islands (Malvinas)"
    FO = "Faroe Islands"
    FJ = "Fiji"
    FI = "Finland"
    FR = "France"
    GF = "French Guiana"
    PF = "French Polynesia"
    TF = "French Southern Territories"
    GA = "Gabon"
    GM = "Gambia"
    GE = "Georgia"
    DE = "Germany"
    GH = "Ghana"
    GI = "Gibraltar"
    GR = "Greece"
    GL = "Greenland"
    GD = "Grenada"
    GP = "Guadeloupe"
    GU = "Guam"
    GT = "Guatemala"
    GG = "Guernsey"
    GN = "Guinea"
    GW = "Guinea-Bissau"
    GY = "Guyana"
    HT = "Haiti"
    HM = "Heard Island and McDonald Islands"
    VA = "Holy See (Vatican City State)"
    HN = "Honduras"
    HK = "Hong Kong"
    HU = "Hungary"
    IS = "Iceland"
    IN = "India"
    ID = "Indonesia"
    XZ = "Installations in International Waters"
    IR = "Iran, Islamic Republic of"
    IQ = "Iraq"
    IE = "Ireland"
    IM = "Isle of Man"
    IL = "Israel"
    IT = "Italy"
    JM = "Jamaica"
    JP = "Japan"
    JE = "Jersey"
    JO = "Jordan"
    KZ = "Kazakhstan"
    KE = "Kenya"
    KI = "Kiribati"
    KP = "Korea, Democratic People's Republic of"
    KR = "Korea, Republic of"
    KW = "Kuwait"
    KG = "Kyrgyzstan"
    LA = "Lao People's Democratic Republic"
    LV = "Latvia"
    LB = "Lebanon"
    LS = "Lesotho"
    LR = "Liberia"
    LY = "Libya"
    LI = "Liechtenstein"
    LT = "Lithuania"
    LU = "Luxembourg"
    MO = "Macao"
    MG = "Madagascar"
    MW = "Malawi"
    MY = "Malaysia"
    MV = "Maldives"
    ML = "Mali"
    MT = "Malta"
    MH = "Marshall Islands"
    MQ = "Martinique"
    MR = "Mauritania"
    MU = "Mauritius"
    YT = "Mayotte"
    MX = "Mexico"
    FM = "Micronesia, Federated States of"
    MD = "Moldova, Republic of"
    MC = "Monaco"
    MN = "Mongolia"
    ME = "Montenegro"
    MS = "Montserrat"
    MA = "Morocco"
    MZ = "Mozambique"
    MM = "Myanmar"
    NA = "Namibia"
    NR = "Nauru"
    NP = "Nepal"
    NL = "Netherlands"
    NC = "New Caledonia"
    NZ = "New Zealand"
    NI = "Nicaragua"
    NE = "Niger"
    NG = "Nigeria"
    NU = "Niue"
    NF = "Norfolk Island"
    MK = "North Macedonia"
    MP = "Northern Mariana Islands"
    NO = "Norway"
    OM = "Oman"
    PK = "Pakistan"
    PW = "Palau"
    PS = "Palestine, State of"
    PA = "Panama"
    PG = "Papua New Guinea"
    PY = "Paraguay"
    PE = "Peru"
    PH = "Philippines"
    PN = "Pitcairn"
    PL = "Poland"
    PT = "Portugal"
    PR = "Puerto Rico"
    QA = "Qatar"
    RE = "Reunion"
    RO = "Romania"
    RU = "Russian Federation"
    RW = "Rwanda"
    BL = "Saint Barthélemy"
    SH = "Saint Helena, Ascension and Tristan Da Cunha"
    KN = "Saint Kitts and Nevis"
    LC = "Saint Lucia"
    MF = "Saint Martin (French Part)"
    PM = "Saint Pierre and Miquelon"
    VC = "Saint Vincent and the Grenadines"
    WS = "Samoa"
    SM = "San Marino"
    ST = "Sao Tome and Principe"
    SA = "Saudi Arabia"
    SN = "Senegal"
    RS = "Serbia"
    SC = "Seychelles"
    SL = "Sierra Leone"
    SG = "Singapore"
    SX = "Sint Maarten (Dutch Part)"
    SK = "Slovakia"
    SI = "Slovenia"
    SB = "Solomon Islands"
    SO = "Somalia"
    ZA = "South Africa"
    GS = "South Georgia and the South Sandwich Islands"
    SS = "South Sudan"
    ES = "Spain"
    LK = "Sri Lanka"
    SD = "Sudan"
    SR = "Suriname"
    SJ = "Svalbard and Jan Mayen"
    SE = "Sweden"
    CH = "Switzerland"
    SY = "Syrian Arab Republic"
    TW = "Taiwan, Province of China"
    TJ = "Tajikistan"
    TZ = "Tanzania, United Republic of"
    TH = "Thailand"
    TL = "Timor-Leste"
    TG = "Togo"
    TK = "Tokelau"
    TO = "Tonga"
    TT = "Trinidad and Tobago"
    TN = "Tunisia"
    TR = "Türkiye"
    TM = "Turkmenistan"
    TC = "Turks and Caicos Islands"
    TV = "Tuvalu"
    UG = "Uganda"
    UA = "Ukraine"
    AE = "United Arab Emirates"
    GB = "United Kingdom"
    US = "United States"
    UM = "United States Minor Outlying Islands"
    UY = "Uruguay"
    UZ = "Uzbekistan"
    VU = "Vanuatu"
    VE = "Venezuela"
    VN = "Viet Nam"
    VG = "Virgin Islands, British"
    VI = "Virgin Islands, U.S."
    WF = "Wallis and Futuna"
    EH = "Western Sahara"
    YE = "Yemen"
    ZM = "Zambia"
    ZW = "Zimbabwe"



from pydantic import BaseModel
from decimal import Decimal
from typing import Literal, Optional, Any, Union

from pydantic import Field

from werk24.models.property.base import W24Property
from werk24.models.property.fraunhofer import W24FraunhoferLine
from werk24.models.typed_model import W24TypedModel


class W24PropertyRefractiveTolerance(W24TypedModel):
    class Config:
        discriminators = ("refractive_tolerance_type",)

    refractive_tolerance_type: Any


class W24PropertyRefractiveToleranceValue(W24PropertyRefractiveTolerance):
    refractive_tolerance_type: Literal["VALUE"] = "VALUE"
    blurb: str = Field(examples=["±0.0005"])
    deviation_upper: Optional[Decimal] = Field(
        examples=[Decimal("0.0005")], default=None)
    deviation_lower: Optional[Decimal] = Field(
        examples=[Decimal("-0.0005")], default=None)


class W24PropertyRefractiveToleranceStep(W24PropertyRefractiveTolerance):
    refractive_tolerance_type: Literal["STEP"] = "STEP"
    blurb: str = Field(examples=["Step 3", "Step 0.5"])
    step: Decimal = Field(examples=[Decimal("3"), Decimal("0.5")])


W24PropertyRefractiveToleranceType = Union[
    W24PropertyRefractiveToleranceValue, W24PropertyRefractiveToleranceStep
]


class W24PropertyRefractiveVariation(BaseModel):
    class Config:
        discriminators = ("variation_type",)

    variation_type: Any


class W24PropertyRefractiveVariationSchottGrade(W24PropertyRefractiveVariation):
    """ Refractive Variation Schott Grade """
    variation_type: Literal["SCHOTT_GRADE"] = "SCHOTT_GRADE"
    blurb: str = Field(examples=("LN", "LH1", "LH2"))
    grade: str = Field(examples=["LN", "LH1", "LH2"])


class W24PropertyRefractiveVariationIso12123(W24PropertyRefractiveVariation):
    """ Refractive Variation ISO 12123 """
    variation_type: Literal["ISO_12123"] = "ISO_12123"
    blurb: str = Field(examples=("NV20", "NV10", "NV05"))
    grade: str = Field(examples=["NV20", "NV10", "NV05"])


class W24PropertyRefractiveVariationFreeText(W24PropertyRefractiveVariation):
    """ Refractive Variation Free Text """
    free_text: str
    variation_type: Literal["FREETEXT"] = "FREETEXT"


W24PropertyRefractiveVariationType = Union[
    W24PropertyRefractiveVariationSchottGrade,
    # W24PropertyRefractiveVariationIso12123
    None,
    W24PropertyRefractiveVariationFreeText,
]


class W24PropertyRefractiveIndex(W24Property):
    """ Parent for all Refractive Index """
    property_type: Literal["REFRACTIVE_INDEX"] = "REFRACTIVE_INDEX"


class W24PropertyRefractiveIndexValue(W24PropertyRefractiveIndex):
    """ Refractive Index Value """
    property_subtype: Literal["VALUE"] = "VALUE"

    blurb: str = Field(
        examples=[
            "nd = 1.72047 ±0.0005",
            "ne = 1.57487 ±0.0005",
        ]
    )

    line: Optional[W24FraunhoferLine] = Field(
        examples=[
            W24FraunhoferLine.D_LINE,
            W24FraunhoferLine.E_LINE,
        ],
        default=None
    )
    value: Decimal = Field(examples=[Decimal("1.72047")])
    tolerance: Optional[W24PropertyRefractiveToleranceType] = Field(
        default=None)
    variation: Optional[W24PropertyRefractiveVariationType] = Field(
        examples=[W24PropertyRefractiveVariationSchottGrade(
            blurb="NV20", grade="NV20")],
        default=None
    )


class W24PropertyRefractiveIndexFreeText(W24PropertyRefractiveIndex):
    """ Refractive Index Free Text """
    free_text: str
    variation_type: Literal["FREETEXT"] = "FREETEXT"


W24PropertyRefractiveIndexType = Union[
    W24PropertyRefractiveIndexValue,
    W24PropertyRefractiveIndexFreeText,
]



from typing import Tuple
import abc
from decimal import Decimal
from enum import Enum
from typing import List, Optional

from pydantic import UUID4, BaseModel

from .base_feature import W24BaseFeatureModel


class W24GDTCharacteristic(str, Enum):
    """ Enum of all possible Characteristics
    following ISO 1101.

    !!! note
        When the GDT frame only describes the location
        of a reference datum, the value DATUM_INDICATOR is used
    """
    FORM_STRAIGHTNESS = "⏤"
    FORM_FLATNESS = "⏥"
    FORM_CIRCULARITY = "○"
    FORM_CYLINDRICITY = "⌭"
    PROFILE_OF_SURFACE = "⌓"
    PROFILE_OF_LINE = "⌒"
    ORIENTATION_PERPENDICULARITY = "⟂"
    ORIENTATION_ANGULARITY = "∠"
    ORIENTATION_PARALLELISM = "∥"
    LOCATION_POSITION = "⌖"
    LOCATION_CONCENTRICITY = "◎"
    LOCATION_SYMMETRY = "⌯"
    RUNOUT_CIRCULAR = "↗"
    RUNOUT_TOTAL = "⌰"
    DATUM_INDICATOR = "[DATUM]"


class W24GDTFeatureAssociated(str, Enum):
    """ Enum of all associated tolerated features
    """
    MINIMAX = "Ⓒ"
    GAUSSIAN = "Ⓖ"
    MIN_CIRCUMSCRIBED = "Ⓝ"
    MAX_CIRCUMSCRIBED = "Ⓧ"
    TANGENT = "Ⓣ"


class W24GDTFeatureDerived(str, Enum):
    """ Enum of all derived features
    """
    PROJECTED = "Ⓟ"
    MEAN = "Ⓐ"


class W24GDTZoneCombination(str, Enum):
    """ Enum of all tolerance zone combinations
    """
    COMBINED = "CZ"
    SEPARATED = "SZ"


class W24GDTZoneShape(str, Enum):
    """ Enum of all zone shapes
    """
    DIAMETER = "Ø"
    DIAMETER_SPHERE = "S⌀"


class W24GDTZoneConstraint(str, Enum):
    """ Enum of the Zone Constraints
    """
    UNSPECIFIED_INCLINATION = "OZ"
    UNSPECIFIED_OFFSET = "VA"
    ORIENTATION_ONLY = "><"


class W24GDTDatum(BaseModel):
    """ Preliminary implementation of the GD&T Datum

    Attributes:
        blurb: Reference name. Typically: A,B,C ...
            Can also contain more complex names, e.g., (A-B-C-D)[CM]

    !!! note
        Future implementations might allow fine-grained
        access to the attributes of complex names:
        e.g., (A-B-C-D)[CM]
    """

    blurb: str


class W24GDTZoneOffset(BaseModel):
    """ Specified offset indicated by leading UZ...

    Attributes:
        blurb: blurb of the specified offset for human consumption.
            e.g., UZ+0.15, UZ-0.2, UZ+0.1:0.2
    """

    blurb: str


class W24GDTZoneValue(BaseModel):
    """ Preliminary defintion of the GDT Zone Value
    Future implementation will give access to the
    width and extend seperately

    Attributes:

        blurb: String representation for human consumption
            e.g., 0.05/12x10°

        width_min: Minimal width. Also used when no maximal width
            is defined.

        width_max: Optional maximal width.

        extend_quantity: Optional quantity of the spacing

        extend_shape: Optional shape of the extend

        extend_value: Optional extend value

        extend_angle: Optional angle of the extend
    """

    blurb: str

    width_min: Decimal

    width_max: Optional[Decimal]

    extend_quantity: Optional[int]

    extend_shape: Optional[W24GDTZoneShape]

    extend_value: Optional[Decimal]

    extend_angle: Optional[Decimal]


class W24GDTFilterType(str, Enum):
    """ Preliminary list of feature filters

    Filters remaining: RG, S, OH, SW, AB, CW
    """
    GAUSSIAN = "G"
    SPLINE = "S"
    SPLINE_WAVELET = "SW"
    COMPLEX_WAVELET = "CW"
    ROBUST_GAUSSIAN = "RG"
    ROBUST_SPLINE = "RS"
    OPENING_BALL = "OB"
    OPENING_HORIZONTAL_SEGMENT = "OH"
    OPENING_DISC = "OC"
    CLOSING_BALL = "CB"
    CLOSING_HORIZONTAL_SEGMENT = "CH"
    CLOSING_DISC = "CD"
    ALTERNATING_BALL = "AB"
    ALTERNATING_HORIZONTAL_SEGMENT = "AH"
    ALTERNATING_DISC = "AD"
    FOURIER = "F"
    HULL = "H"


class W24GDTFilter(BaseModel, abc.ABC):
    """ Abstract base class to describe feature filters

    Attributes:
        blurb: String representation of the file for human consumption

        filter_type: Filter Type to facilitate deserialization
    """

    blurb: str

    filter_type: W24GDTFilterType


# class W24GDTFilterG(W24GDTFilter):
#     """ Gaussian Feature filter without contraints
#     """
#     filter_type = W24GDTFilterType.GAUSSIAN

#     wavelength_min: float
#     """ Minimum wavelength in wavers/revolution
#     """

#     wavelength_max: float
#     """ Maximum wavelength in wavers/revolution
#     """


class W24GDTReferenceAssociation(str, Enum):
    """ Association of the Reference Element
    """

    MINIMAX = "C"
    """ Minimax (Tschebyschew-) Element without contraint
    """

    MINIMAX_EXTERNAL = "CE"
    """ Minimax (Tschebyschew-) Element with external
    constraint
    """

    MINIMAX_INTERNAL = "CI"
    """ Minimal (Tschebyschew-) Element with internal
    constraint
    """

    GAUSSIAN = "G"
    """ Gaussian Least Square Element without constraint
    """

    GAUSSIAN_EXTERNAL = "GE"
    """ Gaussian Least Square Element with external
    constraint
    """

    GAUSSIAN_INTERNAL = "GI"
    """ Gaussian Least Square Element with external
    constraint
    """

    MIN_CIRCUMSCRIBED = "N"
    """ Minimal circumscribed element
    """

    MAX_CIRCUMSCRIBED = "X"
    """ Maximal circumscribed element
    """


class W24GDTReferenceParameter(str, Enum):
    """ Parameter of the reference element
    """

    PEAK_VALUE = "P"
    """ Highest value
    """

    VALLEY_VALUE = "V"
    """ Lowest value
    """

    DEVISATION_SPAN = "T"
    """ Deviation span
    """

    STANDARD_DEVIATION = "Q"
    """ Root mean square deviation
    """


class W24GDTMaterialCondition(str, Enum):
    """ Enum for Material Conditions
    """

    MAXIMUM = "Ⓜ"
    """ Maximum material condition (MMC)
    """

    MINIMUM = "Ⓛ"
    """ Mimimal material condition (LMC)
    """

    RECIPROCITY = "Ⓡ"
    """ Reciprocity
    """


class W24GDTState(str, Enum):
    """ Enum for the State
    """
    FREE = "Ⓕ"


class W24GDTFrame(BaseModel):
    """ Representation of the Geometric Dimensioning
    and Toleration frame

    Attributes:
        gdt_id: Unique id of the GDT

        blurb: String representation of the label for human consumption
            e.g., [⌖|⌀0.3Ⓜ|A|B|C]

        characteristic: Section for geometric characteristic e.g.: ⌓

        zone_shape: Tolerance zone shape, e.g, S⌀

        zone_value: GDT value: e.g., 0.03
            Is optional to support Datum Feature Indicators

        zone_combinations: Ordered list of zone combinations, e.g., CZ, SZ

        zone_offset: Optional specified offset, e.g., UZ-0.2

        zone_constraint: Optional zone constraint: e.g., OZ, VA

        feature_filter: Optional feature filter

        feature_associated: Associated toleraced feature

        feature_derived: Derived Feature

        reference_association: Reference element association

        reference_parameter: Reference element parameter

        material_condition: Material condition

        state: FREE or None

        data: Ordered list of data

    """
    gdt_id: Optional[UUID4] = None

    blurb: str

    characteristic: W24GDTCharacteristic

    zone_shape: Optional[W24GDTZoneShape] = None

    zone_value: Optional[W24GDTZoneValue]

    zone_combinations: List[W24GDTZoneCombination] = []

    zone_offset: Optional[W24GDTZoneOffset] = None

    zone_constraint: Optional[W24GDTZoneConstraint] = None

    feature_filter: Optional[W24GDTFilter] = None

    feature_associated: Optional[W24GDTFeatureAssociated] = None

    feature_derived: Optional[W24GDTFeatureDerived] = None

    reference_association: Optional[W24GDTReferenceAssociation] = None

    reference_parameter: Optional[W24GDTReferenceParameter] = None

    material_condition: Optional[W24GDTMaterialCondition] = None

    state: Optional[W24GDTState] = None

    data: List[W24GDTDatum] = []


class W24GDT(W24BaseFeatureModel):
    """ Parent object for Geometric Dimensionsing and Toleration
    Frames, attaching them to the physical location on the drawing.

    Attributes:
        frame: Representation of the GDT frame

    """
    # !!! DEPRECATED
    bounding_polygon: List[Tuple[float, float]]

    frame: W24GDTFrame

    # measure_label: Optional[W24MeasureLabel] = None
    """ Optional size dimension, typically annotated
    above the feature control frame.

    Example:
        ⌀12.0 + /- 0.1
        [⊥|⌀0.1Ⓜ| A]
    """

    # sectional_plane
    """ Definition of Sectional Plane Indicator will
    follow.
    Example: ⟨[∥| A]
    """  # pylint: disable=pointless-string-statement

    # orientation_plane
    """ Definition of the Orientational Plane Indicator
    will follow
    Example: ⟨[∥| A]⟩
    """  # pylint: disable=pointless-string-statement

    # direction_feature
    """ Defintion of the Direction Feature Indicator
    will follow
    Example: ←[∥| A]
    """  # pylint: disable=pointless-string-statement

    # collection_plane
    """ Defintion of the Collection Plane Indicator
    will follow
    Exmple: ◯[∥| B]
    """  # pylint: disable=pointless-string-statement



from typing import Optional
from werk24.techread_client import LicenseError, W24TechreadClient


def make_client(server:Optional[str] = None) -> W24TechreadClient:
    """
    Make the client.

    This will automatically
    fetch the authentication information
    from the environment variables. We will
    provide you with separate .env files for
    the development and production environments

    Returns:
    -------
    W24TechreadClient: Client instance
    """

    server_wss = None if server is None else f"ws-api.{server}"

    try:
        return W24TechreadClient.make_from_env(
            server_wss=server_wss
        )
    except LicenseError as exception:
        print(f"LICENSE ERROR: {exception}")
        raise SystemExit from exception



from decimal import Decimal
from typing import Optional, Annotated

from pint import Quantity as PintQuantity, UnitRegistry
from pydantic import BaseModel, Field, BeforeValidator, WithJsonSchema, PlainSerializer

from werk24.models.tolerance import W24Tolerance

ureg = UnitRegistry()

Quantity = Annotated[
    PintQuantity,
    BeforeValidator(lambda x: x if isinstance(x, PintQuantity) else ureg(str(x))),
    PlainSerializer(lambda x: str(x), return_type=str),
    WithJsonSchema({"type": "string"}, mode="serialization"),
    WithJsonSchema({"type": "string"}, mode="validation"),
]


class W24PhysicalQuantity(BaseModel):
    """Physical Quantity.

    Physical Quantity with a value, unit and tolerance.
    """

    class Config:
        arbitrary_types_allowed = True

    blurb: str = Field(
        title="blurb",
        description="Blurb of the Physical Property for human consumption.",
    )
    value: Quantity = Field(
        title="value", description="Physical quantity in the string format of Pint."
    )
    tolerance: Optional[W24Tolerance] = None


class W24Value(BaseModel):
    blurb: str
    value: Decimal
    tolerance: Optional[W24Tolerance] = None



""" W24Client Module

DESCRIPTION
    The module contains everything that is needed to
    communicate with the W24 API - allowing you
    to interpret the contents of your technical drawings.

AUTHOR
    Jochen Mattes (Werk24)

EXAMPLE
    # obtain the thumbnail of a page
    drawing_bytes = open(...,"r").read()
    client = W24TechreadClient.make_from_env()
    await client.read_drawing_with_hooks(
        drawing_bytes,
        [Hook(
                ask=W24AskPageThumbnail(),
                function=lambda msg: print("Received Thumbnail of Page")
        ]))
"""
import asyncio
import json
from werk24.crypt import (
    generate_new_key_pair,
    decrypt_with_private_key,
    encrypt_with_public_key,
)
import logging
from io import BufferedReader
import os
from typing import Any
import uuid
from asyncio import iscoroutinefunction
from types import TracebackType
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
    Tuple,
)
from werk24.auth_client import AuthClient

import dotenv
from pydantic import UUID4, BaseModel

from werk24.exceptions import (
    BadRequestException,
    LicenseError,
    RequestTooLargeException,
    ServerException,
    UnsupportedMediaType,
    InsufficientCreditsException
)
from werk24.models.ask import W24Ask
from werk24.models.helpdesk import W24HelpdeskTask
from werk24.models.techread import (
    W24TechreadAction,
    W24TechreadException,
    W24TechreadExceptionLevel,
    W24TechreadExceptionType,
    W24TechreadInitResponse,
    W24TechreadMessage,
    W24TechreadMessageSubtype,
    W24TechreadMessageType,
    W24TechreadRequest,
)
from werk24.techread_client_https import TechreadClientHttps
from werk24.techread_client_wss import TechreadClientWss

# make the logger
logger = logging.getLogger("w24_techread_client")


EXCEPTION_MAP = {
    RequestTooLargeException: W24TechreadExceptionType.DRAWING_FILE_SIZE_TOO_LARGE,
    BadRequestException: W24TechreadExceptionType.DRAWING_FILE_SIZE_TOO_LARGE,
}
""" Map to translate the local exceptions to official
W24Exceptions. This allows us to mock consistent responses
even when the files are rejected before they reach the API
"""

# Default Authentication Region
DEFAULT_AUTH_REGION = "eu-central-1"

# Default Endpoints
DEFAULT_SERVER_WSS = "ws-api.w24.co"
DEFAULT_SERVER_HTTPS = "support.w24.co"

# List of the Locations where we are looking for the license file
# if the user does not specify a path.
LICENSE_LOCATIONS = (".werk24", "werk24_license.txt")

LICENSE_ERROR_TEXT = """
--------------------------------------------------------------------------------

####   ####   #### ########### ##########    ####   #####     ##################
####  #####  ####  ####        ####   #####  ####  ####     #####   #####  #####
 #### ###### ####  ####        ####    ####  #########     ##### ##  ###   #####
 #### ###### ####  ##########  ###########   ########      ######## ### #  #####
  ###### #######   ####        ##########    ##########    ###### ####      ####
  ######  ######   ####        ####   ####   ####  #####   #####     ####  #####
   ####   #####    ########### ####    ####  ####   #####  ###################

--------------------------------------------------------------------------------

General Information
-------------------
Werk24 is a specialized commercial company dedicated to extracting information
from technical documents. The python client you are currently using is designed
to facilitate seamless interaction with our advanced server infrastructure.

License File
------------
We were unable to locate a license file, please schedule a first meeting with
us to learn about the possibilities of our technology:


        >>> https://werk24.io/schedule-a-call?w24cli_license_error <<<


Thank you.
--------------------------------------------------------------------------------
"""


class Hook(BaseModel):
    """
    A Utility class to register callback requests for a specific message_type or W24Ask.

    The 'Hook' object is used for handling and maintaining callback requests. Registering
    an 'ask' should include a complete W24Ask definition, not just the ask type.

    Attributes:
    ----------
    message_type (Optional[W24TechreadMessageType]): Specifies the type of the message.
    message_subtype (Optional[W24TechreadMessageSubtype]): Specifies the subtype of the message.
    ask (Optional[W24Ask]): The complete definition of W24Ask, if any.
    function (Callable): The callback function to be invoked when the resulting information
        is available.

    Note:
    ----
    Either a message_type or an ask must be registered. Be careful when registering an ask;
    a complete W24Ask definition is required, not just the ask type.
    """

    message_type: Optional[W24TechreadMessageType] = None
    message_subtype: Optional[W24TechreadMessageSubtype] = None
    ask: Optional[W24Ask] = None
    function: Callable


class W24TechreadClient:
    """Simple W24Client that allows you to use
    learn more about the content on your Technical
    Drawings.
    """

    def __init__(
        self,
        techread_server_wss: str,
        techread_version: str,
        development_key: str = None,
        support_base_url: str = DEFAULT_SERVER_HTTPS,
        wss_close_timeout: int = 600,
    ):
        """
        Initialize a new W24TechreadClient.

        If you wonder about any of the attributes, have a
        look at the .env file that we provided to you.
        They contain all the information that you will need.

        Args:
        ----
        - techread_server_wss (str): domain name that
            is being used by the websocket client
        - techread_version (str): version that you want to
            connect to
        - development_key (str): key that allows you to submit
            your request to one of the internal architectures.
            You can try guessing or bruteforcing this key;
            we'll just charge you for every request you submit and
            transfer the money to the holiday bonus account.
        """
        self._development_key = development_key

        # Create an empty reference to the authentication
        # service necessary for the Cognito Authentication.
        self._auth_client: Optional[AuthClient] = None

        # HTTP Client
        self._techread_client_https = TechreadClientHttps(
            techread_version,
            support_base_url,
        )

        # WSS Client
        self._techread_client_wss = TechreadClientWss(
            techread_server_wss, techread_version, wss_close_timeout
        )

    async def __aenter__(self) -> "W24TechreadClient":
        """Create the HTTPS and WSS sessions

        Raises:
        ------
        - RuntimeError: Exception is raised if you tried to enter 
            a session before calling the register() method

        Returns:
        -------
        - W24TechreadClient: Version of self with active sessions
        """
        logging.debug("Entering the session")
        await asyncio.gather(
            self._techread_client_https.__aenter__(),
            self._techread_client_wss.__aenter__()
        )
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Ensure that the sessions are closed
        """
        logging.debug("Exiting the session")
        await asyncio.gather(
            self._techread_client_https.__aexit__(exc_type, exc_value, traceback),
            self._techread_client_wss.__aexit__(exc_type, exc_value, traceback)
        )

    def register(
        self,
        cognito_region: Optional[str] = None,
        cognito_identity_pool_id: Optional[str] = None,
        cognito_user_pool_id: Optional[str] = None,
        cognito_client_id: Optional[str] = None,
        cognito_client_secret: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ) -> None:
        """
        Register with the authentication service (i.e., lazy login)

        Note:
        ----
        The Cognito Authentication is only present for backwards
        compatibility. We have moved to a token-based system.
        
        Args:
        ----
        - cognito_region (str): Physical region
        - cognito_identity_pool_id (str): identity pool of W24
        - cognito_client_id (str): the client id of your application
        - cognito_client_secret (str): the client secret of your
            application
        - username (str): the username with which you want to register
        - password (str): the password with which you want to register
        """
        # create an client instance to connect
        # to the authentication service
        self._auth_client = AuthClient(
            cognito_region,
            cognito_identity_pool_id,
            cognito_user_pool_id,
            cognito_client_id,
            cognito_client_secret,
            username,
            password,
            token,
        )
        # ensure that we have a token
        try:
            logging.debug("Authenticating with the authentication service")
            self._auth_client.login()  # type: ignore
        except AttributeError as exc:
            raise RuntimeError(
                "No connection to the authentication service was "
                + "established. Please call register()"
            ) from exc

        # tell the techread clients about it
        self._techread_client_https.register_auth_client(self._auth_client)
        self._techread_client_wss.register_auth_client(self._auth_client)

    @staticmethod
    def generate_encryption_keys(passphrase: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a new RSA key pair and return the private and public key 
        as PEM encoded bytes.

        Args:
        ----
        - passphrase (bytes): The passphrase to encrypt the private key with.

        Returns:
        -------
        - Tuple[bytes, bytes]: The private key and public key as PEM encoded bytes.
        """
        logger.debug("Generating new RSA key pair")
        return generate_new_key_pair(passphrase=passphrase)

    @staticmethod
    def encrypt_with_public_key(public_key_pem: bytes, data: bytes) -> bytes:
        """
        Encrypt the data with the given public key.

        Args:
        ----
        - public_key_pem (bytes): The public key to use for encryption.
        - data (bytes): The data to encrypt.

        Returns:
        -------
        bytes: The encrypted data.
        """
        logger.debug("Encrypting data with public key")
        return encrypt_with_public_key(public_key_pem, data)

    @staticmethod
    def decrypt_with_private_key(
        private_key_pem: bytes, password: bytes, data: bytes
    ) -> bytes:
        """
        Decrypt the data with the given private key.

        Args:
        ----
        private_key_pem (bytes): The private key to use for decryption.
        data (bytes): The data to decrypt.

        Returns:
        -------
        bytes: The decrypted data.
        """
        logger.debug("Decrypting data with private key")
        return decrypt_with_private_key(private_key_pem, password, data)

    async def read_drawing(
        self,
        drawing: Union[BufferedReader, bytes],
        asks: List[W24Ask],
        model: bytes = None,
        max_pages: int = 1,
        drawing_filename: Optional[str] = None,
        sub_account: Optional[UUID4] = None,
        client_public_key_pem: Optional[bytes] = None,
        client_private_key_pem: Optional[bytes] = None,
        client_private_key_passphrase: Optional[bytes] = None,
    ) -> AsyncIterator[W24TechreadMessage]:
        """
        Send a Technical Drawing to the W24 API to read it.

        Args:
        ----
        - drawing (bytes): binary representation of a technical drawing.
            Please refer to the API - documentation to learn which mime
            types are supported.
        - model (bytes): binary representation of the 3d model (typically
            step). This is currently not being used and may come back
            later again.
        - asks (List[W24Ask]): List of Asks that are requested from the API.
            They must derive from the W24Ask object. Refer to the API
            documentation for a full list of supported W24AskTypes
        - max_pages (int): Maximum number of pages that are being processed
            of the submitted file. This protects platform users from
            costly requests caused by a user uploading a single file with
            many pages.
        - drawing_filename (str|None): Optional information about the
            filename of the drawing. Frequently this contains information
            about the drawing id and you can make that information
            available to us through this parameter. If you don't know the
            filename, don't worry, it will still work.
        - sub_account (UUID4|None): Optional specification of the sub-account
            that the request should be attributed to. Sub-accounts allow
            you to manage several customers at the same time and receiving
            separate positions on the monthly invoice.
        - client_public_key_pem (bytes|None): Public key that the server shall
            use to encrypt the callback request. This is only necessary if
            you want to receive the callback in an encrypted form. The
            availability of this feature may depend on your service level.
        - client_private_key_pem (bytes|None): Private key that the server shall
            use to encrypt the callback request. This is only necessary if
            you want to receive the callback in an encrypted form. The
            availability of this feature may depend on your service level.
        - client_private_key_passphrase (bytes|None): Passphrase to decrypt the
            private key. This is only necessary if you want to receive the
            callback in an encrypted form. The availability of this feature
            may depend on your service level.

        Yields:
        ------
        - W24TechreadMessage: Response object obtained from the API
            that indicates the state of your request. Be sure to pass this
            to the read_drawing_listen method

        Raises:
        ------
        - DrawingTooLarge: Exception is raised when the drawing was too
            large to be processed. At the time of writing. The upload
            limit lies at 6 MB (including overhead).
        - UnsupportedMediaType: Exception is raised when the drawing or
            model is submitted in any data type other than bytes.

        """
        # give us some debug information
        logger.debug("API method read_drawing() called")

        # quickly check whether the input type is bytes. If it is string,
        # the presigned-AWS post interestingly returns a 403 error_code
        # without additional information. We want to inform the caller
        # that they submitted the wrong data type.
        # See Github Issue #13
        if not isinstance(drawing, (BufferedReader, bytes)):
            logger.warning("Unsupported media type for drawing")
            raise UnsupportedMediaType(
                "Drawing bytes requires 'bytes' or 'BufferedReader' type"
            )

        # send the initiation command
        init_message, init_response = await self.init_request(
            asks, max_pages, drawing_filename, sub_account
        )
        logger.debug("Init request sent and response received.")

        yield init_message

        # stop if the response is unsuccessful.
        if not init_response.is_successful:
            logger.warning("Init request was not successful.")
            return

        # Start reading the file
        async for message in self.read_request(
            init_response,
            asks,
            drawing,
            model,
            client_public_key_pem,
            client_private_key_pem,
            client_private_key_passphrase,
        ):
            yield message

    async def read_request(
        self,
        init_response: W24TechreadInitResponse,
        asks: List[W24Ask],
        drawing: Union[bytes, BufferedReader],
        model: Optional[bytes] = None,
        client_public_key_pem: Optional[bytes] = None,
        client_private_key_pem: Optional[bytes] = None,
        client_private_key_passphrase: Optional[bytes] = None,
    ) -> AsyncGenerator[W24TechreadMessage, None]:
        """
        Read the request after obtaining the init_response.

        This is helpful when we want to perform the reading
        in the background or in a separate thread.

        Args:
        ----
        - init_response (W24TechreadInitResponse): InitResponse that
            was obtained from the init_request method
        - asks (List[W24Ask]): List of asks that we are asking for.
            This is only used for error handling here.
        - drawing (bytes): Drawing bytes that shall be uploaded
        - model (Optional[bytes], optional): Optional model bytes.
            Defaults to None.
        - client_public_key_pem (Optional[bytes], optional): Public
            key that the server shall use to encrypt the callback
            request. Defaults to None.
        - client_encryption_passphrase (Optional[bytes], optional):
            Passphrase to encrypt the private key. Defaults to None.

        Yields:
        ------
        - W24TechreadMessage: Messages that are received after the
            request was submitted
        """
        logger.debug("API method read_request() called")
        # If the server wants us to encrypt the file, we will do so
        if init_response.public_key is None:
            logger.info(
                "No public key was provided by the server. "
                "Consider upgrading to a higher service level if you need end2end encryption."
            )
            server_public_key = None
        else:
            logger.info("Public key was provided by the server")
            server_public_key = init_response.public_key.encode("utf-8")

        # upload drawing and model. We can do that in parallel.
        # If your user uploads them separately, you could also
        # upload them separately to Werk24.
        try:
            logger.debug("Uploading drawing")
            await self._techread_client_https.upload_associated_file(
                init_response.drawing_presigned_post,
                drawing,
                public_server_key=server_public_key,
            )
            logger.debug("Drawing uploaded")

        # explicitly reraise the exception if the payload is too large
        except (BadRequestException, RequestTooLargeException) as exception:
            async for message in self._trigger_asks_exception(asks, exception):
                yield message
            return

        # Tell Werk24 that all the files have been uploaded
        # correctly and the reading process can be started.
        #
        # NOTE: you will only be able to start the reading
        # process from the websocket connection that
        # initiated the request. If you want to run a
        # stateless-system that separates the initialization
        # from the upload and read stages, you'll need to
        # find a way of handing over the tcp connection :)
        # PS: The AWS API Gateway for websockets might help you
        # here.
        async for message in self._send_command_read(
            client_public_key_pem,
            client_private_key_pem,
            client_private_key_passphrase,
        ):
            yield message

    async def init_request(
        self,
        asks: List[W24Ask],
        max_pages: int,
        drawing_filename: Optional[str],
        sub_account: Optional[UUID4],
    ) -> Tuple[W24TechreadMessage, W24TechreadInitResponse]:
        """
        Initialize a new techread request.

        This method is useful if you want to separate the
        initialization from the upload and read stages.

        This achieves two things:
        1. The server has a couple of 100ms to
           reserves some resources for you, and
        2. The server will create a new request_id
           that you will need when uploading the
           associated files

        Args:
        ----
        - asks (List[W24Ask]): Asks for this request.
        - max_pages (int): Maximum pages to be read.
        - drawing_filename (Optional[str]): Filename of the drawing, if any.
        - sub_account (Optional[UUID4]): Sub-account ID, if any.

        Returns:
        -------
        - Tuple[W24TechreadMessage, W24TechreadInitResponse]: Received
            message and init response.
        """
        logger.debug("API method init_request() called")
        request = W24TechreadRequest(
            asks=asks,
            development_key=self._development_key,
            max_pages=max_pages,
            drawing_filename=drawing_filename,
            sub_account=sub_account,
        )

        await self._techread_client_wss.send_command(
            W24TechreadAction.INITIALIZE.value,
            request.model_dump_json(),
        )
        logger.debug("Techread request submitted")

        message = await self._techread_client_wss.recv_message()
        logger.info("Received request_id %s", message.request_id)

        try:
            response = W24TechreadInitResponse.model_validate(message.payload_dict)
        except ValueError as exception:
            error_message = message.payload_dict.get("message")
            if error_message is not None:
                raise ServerException(error_message) from exception

        return message, response

    async def _send_command_read(
        self,
        client_public_key_pem: Optional[bytes] = None,
        client_private_key_pem: Optional[bytes] = None,
        client_private_key_passphrase: Optional[bytes] = None,
    ) -> AsyncGenerator[W24TechreadMessage, None]:
        """
        Send the request request to the backend
        and yield the resulting messages.

        Yields:
        ------
        - W24TechreadMessage: Receiving messages
        """
        logger.debug("API method _send_command_read() called")
        if client_public_key_pem is not None:
            message = {"public_key": client_public_key_pem.decode("utf-8")}
        else:
            message = {}

        # submit the request the to the API
        logger.debug("Sending techread request")
        await self._techread_client_wss.send_command(
            W24TechreadAction.READ.value, 
            json.dumps(message),
        )
        logger.debug("Techread request submitted")

        # Wait for incoming messages from the server.
        # They will tell you when the individual
        # asks become available. The socket returns
        # strings of jsonified W24TechreadMessage objects.
        #
        # The loop will stop when the websocket is closed
        async for message in self._techread_client_wss.listen():
            logger.debug("Received message %s:%s", message.message_type, message.message_subtype)
            if message.payload_url is not None:
                message.payload_bytes = (
                    await self._techread_client_https.download_payload(
                        message.payload_url,
                        client_private_key_pem,
                        client_private_key_passphrase,
                    )
                )

            # return the message to the caller for immediate
            # consumption
            yield message

    @staticmethod
    async def _trigger_asks_exception(
        asks: List[W24Ask],
        exception_raw: Union[BadRequestException, RequestTooLargeException],
    ) -> AsyncGenerator[W24TechreadMessage, None]:
        """
        Trigger exceptions for all the submitted asks.

        This helps us to mock consistent exception handling
        behavior even when the files are rejected before they
        reach the API.

        Args:
        ----
        - asks (List[W24Ask]): List of all submitted asks
        - exception (RequestTooLargeException): Exception
            that shall be pushed

        Yields:
        ------
        - W24TechreadMessage: Exception message
        """
        logger.debug("API method _trigger_asks_exception() called")

        # get the exception type from the MAP
        try:
            exception_type = EXCEPTION_MAP[type(exception_raw)]

        # if we see an exception that we were not supposed
        # to handle, there must have been a developer passing
        # a new exception type. Let's tell her by rasing
        # a runtime error
        except KeyError as exception:
            raise RuntimeError(
                "Unknown exception type passed: %s" % type(exception_raw)
            ) from exception

        # translate the exception into an official exception
        exception = W24TechreadException(
            exception_level=W24TechreadExceptionLevel.ERROR,
            exception_type=exception_type,
        )

        # then yield one message for each of the requested asks
        for cur_ask in asks:
            yield W24TechreadMessage(
                request_id=uuid.uuid4(),
                message_type=W24TechreadMessageType.ASK,
                message_subtype=cur_ask.ask_type,
                exceptions=[exception],
            )

    @classmethod
    def _get_license_environs(cls, license_path: Optional[str]) -> Dict[str, str]:
        """Get the environment variables
        Where we either select the variables from the license
        files. If that fails we fall back to the true environment
        variables.

        NOTE: We do not want to mix the sources.

        Args:
        ----
        - license_path (Optional[str]): Path of the license files

        Returns:
        -------
        - Dict[str,str]: Key, Value pairs for the environment variables

        Raises:
        ------
        - LicenseError: If the defined license path doesn't exist or if
            relevant environment variable isn't found
        """
        logger.debug("API method _get_license_environs() called")

        # Mimick the old default value of .werk24
        if license_path is None:
            license_path = next(filter(os.path.exists, LICENSE_LOCATIONS), None)
            logger.debug("License path set to %s", license_path)

        # First priority: look for the local license path
        if license_path is not None and os.path.exists(license_path):
            environs_raw = {
                k: v for k, v in dotenv.dotenv_values(license_path).items() if v
            }
        
        # if the caller defined a license path, but it does not exist, raise the exception
        elif license_path is not None:
            logger.warn("License path specified but not valid: %s", license_path)
            raise LicenseError(LICENSE_ERROR_TEXT)

        # Second priority: use the environment variables
        else:
            environs_raw = dict(os.environ)

        return environs_raw

    @classmethod
    def make_from_token(
        cls,
        token: str,
        region: Optional[str] = None,
        server_https: Optional[str] = None,
        server_wss: Optional[str] = None,
        version: str = "v2",
        wss_close_timeout: int = 600,
    ) -> "W24TechreadClient":
        logger.debug("API method make_from_token() called")

        # create a reference to the client
        server_https = server_https or DEFAULT_SERVER_HTTPS
        server_wss = server_wss or DEFAULT_SERVER_WSS
        client = W24TechreadClient(server_wss, version, wss_close_timeout=wss_close_timeout)

        # register the credentials. This will in effect
        # only set the variabels in the authorizer. It will
        # not trigger a network request
        client.register(cognito_region=region, token=token)

        # return the client
        return client

    @classmethod
    def make_from_env(
        cls,
        license_path: Optional[str] = None,
        auth_region: Optional[str] = None,
        server_https: Optional[str] = None,
        server_wss: Optional[str] = None,
        version: str = "v2",
        wss_close_timeout: int = 600,
    ) -> "W24TechreadClient":
        """
        Small helper function that creates a new
        W24TechreadClient from the environment info.

        Args:
        ----
        - license_path (Optional[str]: path to the License file.
            By default we are looking for a .werk24 or werk24_license.txt
            file in the current cwd. If argument is set to None, we are
            not loading any file and relying on the ENVIRONMENT variables only
        - auth_region (Optional[str]): AWS Region of the Authentication
            Service.
            Takes priority over environ W24TECHREAD_AUTH_REGION and
            DEFAULT_AUTH_REGION
        - server_https (Optional[str]): HTTPS endpoint of the Werk24 API.
            Takes priority over environ W24TECHREAD_SERVER_HTTPS and
            DEFAULT_SEVER_HTTPS
        - version (Optional[str]): Version of the Werk24 API.
            Takes priority over environ W24TECHREAD_VERSION and
            DEfAULT_VERSION
        - wss_close_timeout (int): Timeout for the WSS connection.
            Defaults to 600 seconds.

        Raises:
        ------
        - FileNotFoundError: Raised when you pass a path to a license file
            that does not exist
        - UnauthorizedException: Raised when the credentials were not
            accepted by the API

        Returns:
        -------
        - W24TechreadClient: The techread Client
        """
        logger.debug("API method make_from_env() called")

        # get the license variables from the environment variables and
        # the license file.
        environs = cls._get_license_environs(license_path)

        # define a small helper function that finds the first valid
        # value in the supplied list of possible values
        def pick_env(var: Optional[str], env_key: str, default: str) -> str:
            return var or environs.get(env_key, default)

        # then make sure we use the correct priorities
        auth_region = pick_env(
            auth_region, "W24TECHREAD_AUTH_REGION", DEFAULT_AUTH_REGION
        )
        logger.debug("Auth region set to %s", auth_region)

        server_wss = pick_env(
            server_wss, "W24TECHREAD_SERVER_WSS_V2", DEFAULT_SERVER_WSS
        )
        logger.debug("Server WSS set to %s", server_wss)

        # get the variables from the environment and ensure that they
        # are set. If not, raise an exception
        try:
            # create a reference to the client
            client = W24TechreadClient(server_wss, version, wss_close_timeout=wss_close_timeout)

            # register the credentials. This will in effect
            # only set the variabels in the authorizer. It will
            # not trigger a network request
            client.register(
                auth_region,
                environs.get("W24TECHREAD_AUTH_IDENTITY_POOL_ID"),
                environs.get("W24TECHREAD_AUTH_USER_POOL_ID"),
                environs.get("W24TECHREAD_AUTH_CLIENT_ID"),
                environs.get("W24TECHREAD_AUTH_CLIENT_SECRET"),
                environs.get("W24TECHREAD_AUTH_USERNAME"),
                environs.get("W24TECHREAD_AUTH_PASSWORD"),
                environs.get("W24TECHREAD_AUTH_TOKEN"),
            )

        except KeyError:
            raise LicenseError(LICENSE_ERROR_TEXT)

        # return the client
        return client

    async def read_drawing_with_callback(
        self,
        drawing: Union[BufferedReader, bytes],
        asks: List[W24Ask],
        callback_url: str,
        max_pages: int = 5,
        drawing_filename: Optional[str] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        public_key: Optional[bytes] = None,
    ) -> UUID4:
        """
        Read the Drawings and register a callback URL.

        This method is useful if you want to separate the initialization from the 
        upload and read stages.

        You can simply specify the callback URL that shall receive the message responses. 
        This function will return after sending the request to the API. The callback URL 
        will be called asynchronously. Keep in mind that the callback speed depends on your
        service level.

        Args:
        ----
        - drawing (Union[BufferedReader, bytes]):
            Drawing that you want to process
        - asks (List[W24Ask]):
            List of all the information that you want to obtain
        - callback_url (str):
            URL that shall receive the callback requests
        - max_pages (int, optional):
            Maximum number of pages that shall be processed.
            Defaults to 5.
        - drawing_filename (Optional[str], optional):
            Filename of the drawing. Defaults to None.
        - callback_headers (Optional[Dict[str, str]], optional):
            Headers that shall be sent with the callback request. Defaults to None.
        - public_key (Optional[bytes], optional):
            Public key that the server shall use to encrypt the callback request. Defaults to None.
            Note: availability of this feature may depend on your service level.

        Raises:
        ------
        - ServerException: Raised when the server returns an ERROR message
        - InsufficentCreditsException: Raised when the user does not have enough credits 
            to perform the request

        Returns:
        -------
        - UUID4: Request ID of the request
        """
        logger.debug("API method read_drawing_with_callback() called")

        # send the request to the API
        try:
            return await self._techread_client_https.read_drawing_with_callback(
                drawing,
                asks,
                callback_url,
                max_pages=max_pages,
                drawing_filename=drawing_filename,
                callback_headers=callback_headers,
                public_key=public_key,
            )
        except ServerException:
            raise
        except InsufficientCreditsException:
            raise

    async def read_drawing_with_hooks(
        self,
        drawing: Union[BufferedReader, bytes],
        hooks: List[Hook],
        max_pages: int = 5,
        drawing_filename: Optional[str] = None,
        sub_account: Optional[UUID4] = None,
        client_public_key_pem: Optional[bytes] = None,
        client_private_key_pem: Optional[bytes] = None,
        client_private_key_passphrase: Optional[bytes] = None,
    ) -> None:
        """
        Send the drawing to the API (can be PDF or image)
        and register a number of callbacks that are triggered
        once the asks become available.

        Args:
        ----
        - drawing_bytes (bytes): Technical Drawing as Image or PDF
        - hooks (List[Hook]): List of Callback you want to obtain

        Raises:
        ------
        - ServerException: Raised when the server returns an ERROR message
        """
        logger.debug("API method read_drawing_with_hooks() called")

        # filter the callback requests to only contain
        # the ask types
        asks_list = [cur_ask.ask for cur_ask in hooks if cur_ask.ask is not None]
        logger.debug("Filtered asks: %s", asks_list)

        try:
            # send out the request and make a generator
            # that triggers when the result of an ask
            # becomes available
            async for message in self.read_drawing(
                drawing,
                asks_list,
                max_pages=max_pages,
                drawing_filename=drawing_filename,
                sub_account=sub_account,
                client_public_key_pem=client_public_key_pem,
                client_private_key_pem=client_private_key_pem,
                client_private_key_passphrase=client_private_key_passphrase,
            ):
                await self.call_hooks_for_message(message, hooks)

        # explicitly reraise server exceptions
        except ServerException:  # pylint: disable=try-except-raise
            raise

    async def call_hooks_for_message(
        self, message: W24TechreadMessage, hooks: List[Hook]
    ) -> None:
        """
        Call the hook function for the response message.

        Args:
        ----
        - message (W24TechreadMessage): Message returned from the
            read_drawing method
        - hooks (List[Hook]): List of hooks from which we need to
            pick the suited one

        Raises:
        ------
        - ServerException: raised when the server returns an ERROR
            message
        """
        logger.debug("API method call_hooks_for_message() called")
        hook_function = self._get_hook_function_for_message(message, hooks)
        if hook_function is None:
            return

        # if the hook_function is not callable, we want to warn the user,
        # rather than throwing an exception
        if not callable(hook_function):
            logger.warning(
                "You registered a non-callable trigger of type '%s' with "
                "the message_type '%s'. Please make sure that you are using "
                "a Callable (e.g, def or lambda)",
                type(hook_function),
                message.message_type,
            )
            return

        # if everything went well, we call the trigger with
        # the message as payload. Be sure to call the
        # function asymmetrically if supported
        (
            await hook_function(message)
            if iscoroutinefunction(hook_function)
            else hook_function(message)
        )

    @staticmethod
    def _get_hook_function_for_message(
        message: W24TechreadMessage, hooks: List[Hook]
    ) -> Optional[Callable]:
        """
        Get the hook function that corresponds to the message type.

        Args:
        ----
        - message (W24TechreadMessage): Message returned from the read_drawing method
        - hooks (List[Hook]): List of hooks from which we need to pick the suited one

        Returns:
        -------
        - Optional[Callable]: Hook function that should be called
        """
        logger.debug("API method _get_hook_function_for_message() called")

        def hook_filter(hook: Hook) -> bool:
            if message.message_type == W24TechreadMessageType.ASK:
                return (
                    hook.ask is not None
                    and message.message_subtype.value == hook.ask.ask_type.value
                )
            else:
                return (
                    hook.message_type is not None
                    and hook.message_subtype is not None
                    and message.message_type == hook.message_type
                    and message.message_subtype == hook.message_subtype
                )

        # return the first positive case
        for cur_hook in filter(hook_filter, hooks):
            return cur_hook.function

        # if we are still here, we have an unknown message type, which
        # probably is being caused by an API update. We want to ensure
        # that the user is being informed, but we do not want to break
        # the existing functionality -> warning
        logger.debug(
            "Ignoring message of type %s:%s - no hook registered",  # noqa
            message.message_type,
            message.message_subtype,
        )
        return None

    async def create_helpdesk_task(self, task: W24HelpdeskTask) -> W24HelpdeskTask:
        """
        Create a Helpdesk ticket.

        Args:
        ----
        - task (W24HelpdeskTask): Helpdesk task to be created

        Raises:
        ------
        - BadRequestException: Raised when the request body
            cannot be interpreted. This normally indicates
            that the API version has been updated and that
            we missed a corner case. If you encounter this
            exception, it is very likely our mistake. Please
            get in touch!
        - UnauthorizedException: Raised when the token
            or the requested file have expired
        - ResourceNotFoundException: Raised when you are requesting
            an endpoint that does not exist. Again, you should
            not encounter this, but if you do, let us know.
        - RequestTooLargeException: Raised when the status
            code was 413
        - UnsupportedMediaTypException: Raised when the file you
            submitted cannot be read(because its media type
            is not supported by the API).
        - ServerException: Raised for all other status codes
            that are not 2xx

        Returns:
        -------
        - W24HelpdeskTask: Created helpdesk task with an updated task_id.
        """
        logger.debug("API method create_helpdesk_task() called")
        return await self._techread_client_https.create_helpdesk_task(task)



from enum import Enum
from pydantic import BaseModel
from typing import Optional

from werk24.models.fraction import W24Fraction


class W24ViewType(str, Enum):
    """An enumeration of view types.
    """
    VIEW = "VIEW"
    SECTIONAL = "SECTIONAL"
    DETAIL = "DETAIL"
    ISOMETRIC = "ISOMETRIC"


class W24View(BaseModel):
    """Representation of a View Name

    Attributes:
    ----------
    view_type (W24ViewType): The view type, an enumeration member.
    blurb (str): The blurb of the view, a string.
    name (str): Name of the view
    scale (W24Fraction): The scale of the view, a fraction represented
        by the W24Fraction model.
    """
    view_type: W24ViewType
    blurb: str
    name: str
    scale: Optional[W24Fraction]



from typing import Literal, Union

from pydantic import Field

from werk24.models.property.base import W24Property
from werk24.models.value import W24PhysicalQuantity, ureg


class W24PropertyStressBirefringence(W24Property):
    """Parent for all Glass Homogeneity Properties"""

    property_type: Literal["STRESS_BIREFRINGENCE"] = "STRESS_BIREFRINGENCE"


class W24PropertyStressBirefringenceIso10110Value(W24PropertyStressBirefringence):
    """ISO 10110 Stress Birefringence Value"""

    property_subtype: Literal["ISO_10110_VALUE"] = "ISO_10110_VALUE"
    blurb: str = Field(examples=["0/8"])
    value: W24PhysicalQuantity = Field(
        examples=[W24PhysicalQuantity(
            blurb="8nm/cm", value=8 * ureg.nm / ureg.cm)]
    )


class W24PropertyStressBirefringenceIso10110Grade(W24PropertyStressBirefringence):
    """ISO 10110 Stress Birefringence Grade"""

    property_subtype: Literal["ISO_10110_GRADE"] = "ISO_10110_GRADE"
    blurb: str = Field(examples=["SB20", "SB02"])
    grade: str = Field(examples=["SB20", "SB02"])


class W24PropertyStressBirefringenceFreeText(W24PropertyStressBirefringence):
    """Stress Birefringence Free Text"""
    free_text: str
    variation_type: Literal["FREETEXT"] = "FREETEXT"


W24PropertyStressBirefringenceType = Union[
    W24PropertyStressBirefringenceIso10110Value,
    W24PropertyStressBirefringenceIso10110Grade,
    W24PropertyStressBirefringenceFreeText,
]



from enum import Enum

class W24ProjectionMethod(str, Enum):
    """Projection Method according to ISO 128
    """
    FIRST_ANGLE_PROJECTION = "FIRST_ANGLE_PROJECTION"
    THIRD_ANGLE_PROJECTION = "THIRD_ANGLE_PROJECTION"


""" Command Line Interface for W24 Techread
"""
import os, tempfile
import subprocess
import platform
import traceback
from datetime import datetime
import argparse
from termcolor import colored
from colorama import just_fix_windows_console
import io
import json
import logging
from werk24.exceptions import TechreadException
import os
from collections import namedtuple
from typing import Any, Dict, List, Optional

from dotenv import load_dotenv
import random
import string
from werk24.cli import utils
from werk24.models.ask import (
    W24AskCanvasThumbnail,
    W24AskPageThumbnail,
    W24AskPartFamilyCharacterization,
    W24AskSectionalThumbnail,
    W24AskSheetAnonymization,
    W24AskSheetThumbnail,
    W24AskTitleBlock,
    W24AskVariantCAD,
    W24AskVariantExternalDimensions,
    W24AskVariantGDTs,
    W24AskVariantMeasures,
    W24AskVariantRadii,
    W24AskDebug,
    W24AskVariantRoughnesses,
    W24AskVariantThreadElements,
    W24AskVariantProcesses,
)

from werk24.models.techread import (
    W24TechreadException,
    W24TechreadMessageSubtypeError,
    W24TechreadMessageSubtypeProgress,
    W24TechreadMessageType,
    W24TechreadMessage,
)
from werk24.techread_client import LICENSE_LOCATIONS, Hook

# fix the windows console color issue
just_fix_windows_console()

def random_string(length: int) -> str:
    """
    Generate a random string of a given length

    Args:
    ----
    - length (int): Length of the random string

    Returns:
    -------
    - str: Random string of the given length
    """
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))

# load the environment variables
for c_location in LICENSE_LOCATIONS:
    if os.path.exists(c_location):
        load_dotenv(c_location)
        break

# set the log level to info for the test setting
# We recommend using logging.WARNING for production
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",  # noqa
    datefmt="%Y-%m-%d %H:%M:%S",
)

# get the local logger
logger = logging.getLogger(__name__)  # pylint:disable = invalid-name


def _save_and_open_debug(payload_bytes:bytes, extension:str) -> None:
    """
    Save the debug payload to a temporary file and open it
    with the default application for the file type

    This function is required because we need to support
    E2E encryption. So opening the file directly is not
    possible.
    
    Args:
    ----
    payload_bytes (bytes): The payload
    """
    # Create a temporary file that will be deleted after the block exits
    with tempfile.NamedTemporaryFile(delete=False, suffix=f".{extension}") as tmp:
        tmp.write(payload_bytes)
        tmp.flush()  # Ensure all data is written to disk
        tmp_name = tmp.name

    # Determine the command based on the operating system
    system_name = platform.system()
    if system_name == "Windows":
        os.startfile(tmp_name)
    elif system_name == "Darwin":  # macOS
        subprocess.run(["open", tmp_name], check=True)
    elif system_name == "Linux":
        subprocess.run(["xdg-open", tmp_name], check=True)
    else:
        raise OSError(f"Unsupported OS: {system_name}")
    input("Press Enter to delete temporary file...")

    # Ensure the temporary file is deleted
    if os.path.exists(tmp_name):
        try:
            os.remove(tmp_name)
        except Exception as cleanup_error:
            print(f"Failed to delete temporary file: {cleanup_error}")



# make the configuration of what hooks we want to handle and how
HookConfig = namedtuple("HookConfig", "arg ask function")
hook_config = [
    HookConfig(
        "ask_page_thumbnail",
        W24AskPageThumbnail,
        lambda m: _show_image("Page Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sheet_thumbnail",
        W24AskSheetThumbnail,
        lambda m: _show_image("Sheet Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sheet_anonymization",
        W24AskSheetAnonymization,
        lambda m: _show_image("Sheet Anonymization", m.payload_bytes),
    ),
    HookConfig(
        "ask_canvas_thumbnail",
        W24AskCanvasThumbnail,
        lambda m: _show_image("Canvas Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_sectional_thumbnail",
        W24AskSectionalThumbnail,
        lambda m: _show_image("Sectional Thumbnail", m.payload_bytes),
    ),
    HookConfig(
        "ask_variant_external_dimensions",
        W24AskVariantExternalDimensions,
        lambda m: _print_payload("Ask Variant Ext. Dimensions", m),
    ),
    HookConfig(
        "ask_variant_gdts",
        W24AskVariantGDTs,
        lambda m: _print_payload("Ask Variant GDTs", m),
    ),
    HookConfig(
        "ask_variant_measures",
        W24AskVariantMeasures,
        lambda m: _print_payload("Ask Variant Measures", m),
    ),
    HookConfig(
        "ask_variant_radii",
        W24AskVariantRadii,
        lambda m: _print_payload("Ask Variant Radii", m),
    ),
    HookConfig(
        "ask_variant_roughnesses",
        W24AskVariantRoughnesses,
        lambda m: _print_payload("Ask Variant Roughnesses", m),
    ),
    HookConfig(
        "ask_variant_processes",
        W24AskVariantProcesses,
        lambda m: _print_payload("Ask Variant Processes", m),
    ),
    HookConfig(
        "ask_variant_cad",
        W24AskVariantCAD,
        lambda m: _store_variant_cad(m.payload_dict, m.payload_bytes, m.exceptions),
    ),
    HookConfig(
        "ask_titleblock",
        W24AskTitleBlock,
        lambda m: _print_payload("Ask TitleBlock", m),
    ),
    HookConfig(
        "ask_variant_thread_elements",
        W24AskVariantThreadElements,
        lambda m: _print_payload("Ask Variant Thread Elements", m),
    ),
]


def _store_variant_cad(
    payload_dict: Dict[str, Any],
    payload_bytes: Optional[bytes],
    exceptions: List[W24TechreadException],
) -> None:
    """Store the CAD file the current directory

    Args:
    ----
    payload_dict (Dict[str, Any]): Payload Dictionary
    payload_bytes (bytes): CAD that we received as response
    """
    logger.info(f"Ask Variant CAD\n{payload_dict}")

    # print potential exceptions
    if exceptions:
        _log_exceptions(exceptions)

    # store the payload
    if payload_bytes is not None:
        _store_variant_cad_payload(payload_dict, payload_bytes)


def _log_exceptions(exceptions: List[W24TechreadException]) -> None:
    """Log the encountered exceptions as warnings

    Args:
    ----
    exceptions (List[W24TechreadException]): List of
        encountered exceptions. Can be an empty list.
    """

    for cur_exception in exceptions:
        level = str(cur_exception.exception_level.value)
        exception_type = cur_exception.exception_type.value
        message = f"{level}: {exception_type}"
        logger.warn(message)


def _store_variant_cad_payload(
    payload_dict: Dict[str, Any], payload_bytes: bytes
) -> None:
    """Store the CAD file with the knowledge that
    the payload exists

    Args:
    ----
    payload_dict (Dict[str, Any]): Payload Dictionary
    payload_bytes (bytes): CAD response
    """
    # make the filename
    variant_id = payload_dict.get("variant_id")
    filename = f"./w24_ask_variant_cad_{variant_id}.dxf"

    # and write the content
    with open(filename, "wb+") as file_handle:
        file_handle.write(payload_bytes)

    # tell the user
    logger.info(f"CAD response stored in {filename}")


def _get_drawing(file_path: str) -> Optional[io.BufferedReader]:
    """Get the bytes of the file that shall be read.

    Args:
    ----
    file_path (str): path to the input file

    Returns:
    -------
    io.BufferedReader: file handle to the input file
    """
    try:
        return open(file_path, "rb")
    except FileNotFoundError:
        print(f"File {file_path} not found")
        return None


def _print_payload(log_text: str, message: W24TechreadMessage) -> None:
    """Display the payload dictionary in a format that
    is easy for humans to read

    Args:
    ----
    log_text (str): Headline info
    payload_dict (Dict[str, Any]): Payload dictionary
    """
    print(log_text)
    if message.exceptions:
        _log_exceptions(message.exceptions)
    payload_json = json.dumps(message.payload_dict, indent=4)
    print(payload_json)


def _show_image(log_text: str, image_bytes: bytes) -> None:
    """Display a debug image. The function relies on
    the PIL functionality of Image.show(). The
    actual behavior will thus be different across platforms

    Args:
    ----
    log_text (str): Log Text that we want to write back
        to the command line interface
    image_bytes (bytes): byte representation of the image
        that is to be displayed
    """

    # Import the PIL at runtime to make sure that
    # you can still use the client without the need
    # to install Pillow
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
    except ImportError:
        logger.warning(
            "Viewing image-like responses requires the installation"
            " of the PIL package: pip install pillow. Image skipped."
        )
        return

    # otherwise print the text
    logger.info(log_text)

    # and show the image
    image = Image.open(io.BytesIO(image_bytes))
    try:
        image.show(title=log_text)
    except BaseException:
        logger.warning(
            "Image cannot be displayed. " "Please check your local security settings."
        )
        return


async def main(args: argparse.Namespace) -> None:
    """
    Run the Techread command with the CLI arguments

    Args:
    ----
    args(argparse.Namespace): CLI args generated by
        argparse
    """
    try:
        # make the hooks from the arguments
        hooks = _make_hooks_from_args(args)

        # get the client instance and handle
        # potential errors
        client = utils.make_client(args.server)

        # get the drawing
        drawing = _get_drawing(args.input_file)
        if drawing is None:
            return

        # make a key pair
        passphrase = random_string(16)
        public_key_pem, private_key_pem = client.generate_encryption_keys(
            passphrase=passphrase
        )

        async with client as session:
            await session.read_drawing_with_hooks(
                drawing,
                hooks,
                sub_account=None,
                client_public_key_pem=public_key_pem,
                client_private_key_pem=private_key_pem,
                client_private_key_passphrase=passphrase,
            )
            drawing.close()

    except TechreadException as exception:
        print()
        print(colored("=" * 3 + "[Techread Error]" + "=" * 61, "red"))

        print(colored(exception.cli_message_header, "red"))
        print("-" * 80)

        print(colored(exception.cli_message_body, "yellow"))
        print(colored("=" * 80, "red"))
        raise

    except BaseException as exception:
        catch_crash(exception)


def catch_crash(exception: BaseException) -> None:
    """
    Catch crashes during program execution and log them in a file.

    Args:
    ----
    exception (BaseException): The exception instance captured during crash.

    Returns:
    -------
    None
    """
    filename = write_crash_log()
    print_crash_message(filename)


def write_crash_log() -> str:
    """
    Create a log entry of an internal error that occurred during program execution.

    Returns:
    -------
    filename (str): The name of the log file where the error details are written.
    """

    timestamp = str(datetime.now().isoformat()).replace(":", "-")
    filename = f"werk24-crash-{timestamp}.log"
    content = f"=== [Internal Error] ===\n\nCALLSTACK:\n{traceback.format_exc()}"

    with open(filename, "w") as crash_log:
        crash_log.write(content)

    return filename


def print_crash_message(filename: str) -> None:
    """
    Print an error message regarding an internal crash.

    Args:
    ----
    filename (str): The name of the log file containing the error details.

    Returns:
    -------
    None
    """
    separator_line = colored("=" * 80, "red")

    crash_message = [
        separator_line,
        "",
        "=== [Internal Error] ===",
        "An internal error occurred.",
        "To report the error, please send the file called:",
        "",
        f"  --> {filename} <-- ",
        "",
        "to info@werk24.io",
        "We apologize for the inconvenience.",
        "",
        separator_line,
    ]
    print("\n".join(colored(li, "red") for li in crash_message))


def _make_hooks_from_args(args: argparse.Namespace) -> List[Hook]:
    # add a general hook to deal with internal errors
    hook_error_internal = Hook(
        message_type=W24TechreadMessageType.ERROR,
        message_subtype=W24TechreadMessageSubtypeError.INTERNAL,
        function=lambda msg: logger.error("Internal Error %s", msg.payload_dict),
    )

    # add a general hook to deal with PROGRESS messages
    hook_progress_started = Hook(
        message_type=W24TechreadMessageType.PROGRESS,
        message_subtype=W24TechreadMessageSubtypeProgress.STARTED,
        function=lambda msg: logger.info("Techread process started"),
    )

    # tell the api what asks you are interested in,
    # and define what to do when you receive the result
    hooks = [
        Hook(ask=c.ask(), function=c.function)
        for c in hook_config
        if getattr(args, c.arg)
    ] + [hook_error_internal, hook_progress_started]

    # take special care of teh W24AskPartFamilyCharacterization
    if args.part_family_id is not None:
        c_hook = Hook(
            ask=W24AskPartFamilyCharacterization(part_family_id=args.part_family_id),
            function=lambda m: _print_payload(
                "Part Family Characterization", m.payload_dict
            ),
        )
        hooks.append(c_hook)

    if args.debug_key is not None:
        c_hook = Hook(
            ask=W24AskDebug(debug_key=args.debug_key),
            function=lambda m: _save_and_open_debug(m.payload_bytes, "zip"),
        )
        hooks.append(c_hook)

    return hooks



import os
import io
from typing import Optional, Union
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from typing import Tuple


def generate_new_key_pair(
    passphrase: bytes,
    public_exponent=65537,
    key_size=2048,
) -> Tuple[bytes, bytes]:
    """
    Generate a new RSA key pair and return the private and public key as PEM encoded bytes.

    Args:
    ----
    passphrase: bytes
        The passphrase to encrypt the private key with.

    public_exponent: int
        The public exponent to use for the key pair. Default is 65537.

    key_size: int
        The size of the key in bits. Default is 2048.

    Returns:
    -------
    tuple[bytes, bytes]
        The private key and public key as PEM encoded bytes.
    """
    if isinstance(passphrase, str):
        passphrase = passphrase.encode("utf-8")

    private_key = rsa.generate_private_key(
        public_exponent=public_exponent,
        key_size=key_size,
        backend=default_backend(),
    )

    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.BestAvailableEncryption(passphrase),
    )

    public_key = private_key.public_key()

    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return public_key_pem, private_key_pem


def encrypt_with_public_key(
    public_key_pem: bytes,
    data: Union[bytes, io.BufferedReader],
) -> bytes:
    """
    Encrypt the data with the given public key.

    Args:
    ----
    public_key_pem: bytes
        The public key to use for encryption.

    data: bytes
        The data to encrypt.

    Returns:
    -------
    bytes
        The encrypted data.
    """

    # Ensure that the public key is in bytes
    if isinstance(public_key_pem, str):
        public_key_pem = public_key_pem.encode("utf-8")

    # Load the recipient's public key
    public_key = serialization.load_pem_public_key(
        public_key_pem,
        backend=default_backend(),
    )

    # ensure that the data is in bytes
    if hasattr(data, "read"):
        raw_data = data.read()
    else:
        raw_data = data

    # Encrypt the file content with AES key in GCM mode
    aes_key = os.urandom(32)  # 256-bit AES key
    iv = os.urandom(12)  # GCM standard IV size is 96 bits (12 bytes)
    cipher = Cipher(algorithms.AES(aes_key), modes.GCM(iv), backend=default_backend())
    encryptor = cipher.encryptor()
    encrypted_data = encryptor.update(raw_data) + encryptor.finalize()

    # Encrypt the AES key with the recipient's RSA public key
    encrypted_aes_key = public_key.encrypt(
        aes_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    # Combine the encrypted AES key, IV, tag, and encrypted data
    encrypted_package = encrypted_aes_key + iv + encryptor.tag + encrypted_data
    return encrypted_package


def decrypt_with_private_key(
    private_key_pem: bytes,
    password: Optional[str],
    encrypted_package: bytes,
) -> bytes:
    """
    Decrypt the encrypted package with the given private key.

    Args:
    ----
    private_key_pem: bytes
        The private key to use for decryption.

    passphrase: str | None
        The master key to decrypt the private key.

    encrypted_package: bytes
        The encrypted package to decrypt.

    Returns:
    -------
    bytes
        The decrypted data.
    """
    if isinstance(private_key_pem, str):
        private_key_pem = private_key_pem.encode("utf-8")

    if isinstance(password, str):
        password = password.encode("utf-8")

    # Load the recipient's private key
    private_key = serialization.load_pem_private_key(
        private_key_pem,
        password=password,
        backend=default_backend(),
    )

    # Extract the encrypted AES key, IV, tag, and encrypted data
    encrypted_aes_key = encrypted_package[: private_key.key_size // 8]
    iv = encrypted_package[private_key.key_size // 8 : private_key.key_size // 8 + 12]
    tag = encrypted_package[
        private_key.key_size // 8 + 12 : private_key.key_size // 8 + 28
    ]
    encrypted_data = encrypted_package[private_key.key_size // 8 + 28 :]

    # Decrypt the AES key with the recipient's RSA private key
    aes_key = private_key.decrypt(
        encrypted_aes_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    # Decrypt the file content with the decrypted AES key in GCM mode
    cipher = Cipher(
        algorithms.AES(aes_key),
        modes.GCM(iv, tag),
        backend=default_backend(),
    )
    decryptor = cipher.decryptor()
    decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()

    return decrypted_data



from decimal import Decimal
from typing import Optional

from pydantic import BaseModel

from .size import W24Size
from .depth import W24Depth


class W24CounterBore(BaseModel):
    """Size of the Counterbore including its angle

    Attributes:
        blurb: Blurb for human consumption
        size: Size of the Counterbore as referred in the hole callout.
        depth: Depth of the Counterbore
    """

    blurb: str
    size: W24Size
    depth: W24Depth


class W24CounterDrill(BaseModel):
    """Size of the Counter Drill including its angle

    Attributes:
        blurb: Blurb for human consumption
        size: Size of the Counterdrill as referred in the hole callout.
        depth: Depth of the Counterdrill
        angle: Angle of the Counterdrill in Degrees. Angle of counterdrill
            is preferred to be annotated in a sectional view of the hole.
            As a result, angle is optional and set to None by default.
    """

    blurb: str
    size: W24Size
    depth: W24Depth
    angle: Optional[Decimal] = None


class W24CounterSink(BaseModel):
    """Size of the Countersink including its angle

    Attributes:
        blurb: Blurb for human consumption
        size: Size of the Countersink as referred in the hole callout.
        angle: Angle of the Countersink in Degrees
    """

    blurb: str
    size: W24Size
    angle: Decimal



from typing import Optional, Any

from pydantic import BaseModel

from werk24.models.standard import W24Standard


class W24Feature(BaseModel):
    """BaseModel for all Werk24 Features.

    Attributes:
        blurb (str): Description of the feature
            following the logic of the referenced
            standard. If no standard is references,
            we fall back to the ISO way of doing things.
            See ISO 19136-1:2020.

        standard (Optional[W24Standard]): if available,
            the standard that defines the feature.
    """

    blurb: str
    feature_type: Any
    feature_subtype: Any = None
    standard: Optional[W24Standard] = None



from typing import Any

from werk24.models.typed_model import W24TypedModel


class W24Property(W24TypedModel):
    class Config:
        discriminators = ('property_type', 'property_subtype')

    property_type: Any
    property_subtype: Any
    blurb: str





# from decimal import Decimal
# from enum import Enum
# from typing import Optional

# from pydantic import BaseModel


# class W24Dimension(str, Enum):
#     """Dimensions for the Physical Quantities.

#     NOTE: This list might grow as we add additional
#     properties to our database.
#     """
#     # start with the SI base dimensions
#     LENGTH = "LENGTH"  # [m]
#     MASS = "MASS"  # [kg]
#     TIME = "TIME"  # [s]
#     ELECTRIC_CURRENT = "ELECTRIC_CURRENT"  # [A]
#     THERMODYNAMIC_TEMPERATURE = "THERMODYNAMIC_TEMPERATURE"  # [K]
#     AMOUNT_OF_SUBSTANCE = "AMOUNT_OF_SUBSTANCE"  # [mol]
#     LUMINOUS_INTENSITY = "LUMINOUS_INTENSITY"  # [cd]

#     # derived SI dimensions
#     AREA = "AREA"  # [m2]
#     VOLUME = "VOLUME"  # [m2]
#     VELOCITY = "VELOCITY"  # [m/s]
#     ACCELERATION = "ACCELERATION"  # [m/s2]
#     FORCE = "FORCE"  # [N]
#     PRESSURE = "PRESSURE"  # [Pa]
#     ENERGY = "ENERGY"  # [J]
#     POWER = "POWER"  # [W]
#     ELECTRIC_CHARGE = "ELECTRIC_CHARGE"  # [C]
#     ELECTRIC_POTENTIAL = "ELECTRIC_POTENTIAL"  # [V]
#     ELECTRIC_RESISTANCE = "ELECTRIC_RESISTANCE"  # [OHM]
#     MAGNETIC_FLUX = "MAGNETIC_FLUX"  # [Wb]
#     MAGNETIC_FIELD = "MAGNETIC_FIELD"  # [T]

#     # commonly used derived dimensions
#     LINEAR_DENSITY = "LINEAR_DENSITY"  # [kg/m]
#     AREA_DENSITY = "AREA_DENSITY"  # [kg/m2]
#     MASS_DENSITY = "MASS_DENSITY"  # [kg/m3]


# class W24UnitPrefix(Decimal, Enum):
#     """List of the Unit Prefixes in the SI system.

#     NOTE: the Chinese and Indian systems are based
#     on multiples of 10'000. There are other numbering
#     systems that work on multiples of 12 or 6.
#     """
#     YOTTA = Decimal("10")**24
#     ZETTA = Decimal("10")**21
#     EXA = Decimal("10")**18
#     PETA = Decimal("10")**15
#     TERA = Decimal("10")**12
#     GIGA = Decimal("10")**9
#     MEGA = Decimal("10")**6
#     KILO = Decimal("10")**3
#     HECTO = Decimal("10")**2
#     DECA = Decimal("10")**1
#     DECI = Decimal("10")**-1
#     CENTI = Decimal("10")**-2
#     MILLI = Decimal("10")**-3
#     MICRO = Decimal("10")**-6
#     NANO = Decimal("10")**-9
#     PICO = Decimal("10")**-12
#     FEMTO = Decimal("10")**-15
#     ATTO = Decimal("10")**-18
#     ZEPTO = Decimal("10")**-21
#     YOCTO = Decimal("10")**-24


# class W24Unit(BaseModel):
#     """Physical Unit.

#     Attributes:
#         dimension (W24Dimension): Dimension that is
#             measured by the unit.
#         name (str): Name of the unit.
#         symbol (str): Symbol of the unit. E.g. `l` for liter.
#         unit_prefix (Optional[W24UnitPrefix]): Optional Unit
#             prefix. For example: milli or Mega.
#         unit_power (int): Power of the unit. Eg. 3 for volumes.
#         si_multiple (Decimal): Factor by which the value
#             needs to be multiplied to obtain the SI value of
#             the dimension.

#     """
#     dimension: W24Dimension
#     name: str
#     symbol: str
#     unit_prefix: Optional[W24UnitPrefix]
#     unit_power: int = 1
#     si_factor: Decimal


# class W24UnitFactory:
#     """Factory to create the Units.
#     """

#     @classmethod
#     def _make(
#         cls,
#         dimension: W24Dimension,
#         config_options,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix],
#         unit_power: int = 1
#     ) -> W24Unit:

#         # get the configuration.
#         config = config_options.get(name)
#         if config is None:
#             raise ValueError()

#         return W24Unit(
#             dimension=dimension,
#             unit_prefix=unit_prefix,
#             unit_power=unit_power,
#             symbol=config[0],
#             name=name,
#             si_factor=Decimal(config[1])
#         )

#     @classmethod
#     def make_length(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "meter": ("m", "1"),
#             "inch": ("in",  "0.0254"),
#             "foot": ("ft",  "0.3048"),
#         }
#         return cls._make(W24Dimension.LENGTH, config, name, unit_prefix)

#     @classmethod
#     def make_area(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "square meter": ("m2",  "1"),
#             "square inch": ("in2",  "0.0254"),
#             "square foot": ("ft2",  "0.3048"),
#         }
#         return cls._make(W24Dimension.AREA, config, name, unit_prefix, 2)

#     @classmethod
#     def make_volume(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "cubic meter": ("m3", "1"),
#             "cubic inch": ("in3", "0.0254"),
#             "cubic foot": ("ft3", "0.3048"),
#             "liter": ("l", "0.001"),
#         }
#         return cls._make(W24Dimension.VOLUME, config, name, unit_prefix, 3)

#     @classmethod
#     def make_time(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         # NOTE: a day can sometimes be shorter than 86400 seconds;
#         # but as measurement unit, we'll accept this.
#         config = {
#             "day": ("day", "86400"),
#             "hour": ("h", "3600"),
#             "minute": ("min",  "60"),
#             "second": ("s",  "1"),
#         }
#         return cls._make(W24Dimension.TIME, config, name, unit_prefix, 1)

#     @classmethod
#     def make_mass(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "tonne": ("t", "1000"),
#             "gram": ("g", "0.001"),
#             "pound": ("lb", "0.453592"),
#             "ounce": ("oz",  "0.0283495"),
#             "stone": ("oz",  "6.35029"),
#             "slug": ("slug", "14.5939"),
#             "carat": ("ct", "0.0002"),
#         }
#         return cls._make(W24Dimension.MASS, config, name, unit_prefix, 1)

#     @classmethod
#     def make_linear_density(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "tonne per meter": ("t/m", "1000"),
#             "gram per meter": ("g/m", "0.001"),
#             "pound per meter": ("lb/m", "0.453592"),
#             "ounce per meter": ("oz/m",  "0.0283495"),
#             "stone per meter": ("oz/m",  "6.35029"),
#             "slug per meter": ("slug/m", "14.5939"),
#             "carat per meter": ("ct/m", "0.0002"),

#             "tonne per inch": ("t/in", "39370.1"),
#             "gram per inch": ("g/in", "0.0393701"),
#             "pound per inch": ("lb/in", "17.858"),
#             "ounce per inch": ("oz/in",  "1.11612"),
#             "stone per inch": ("oz/in",  "250.012"),
#             "slug per inch": ("slug/in", "574.563"),
#             "carat per inch": ("ct/in", "0.00787402"),
#         }
#         return cls._make(W24Dimension.LINEAR_DENSITY, config, name, unit_prefix, 1)

#     @classmethod
#     def make_force(
#         cls,
#         name: str,
#         unit_prefix: Optional[W24UnitPrefix] = None
#     ) -> W24Unit:
#         config = {
#             "dyne": ("dyn", "0.00001"),
#             "newton": ("N", "1"),
#             "kilogram-force": ("kgf", "9.806650")
#         }
#         return cls._make(W24Dimension.FORCE, config, name, unit_prefix, 1)


# class W24PhysicalQuantity(BaseModel):
#     value: Decimal
#     unit: W24Unit

#     def convert_to(self, new_unit: W24Unit):

#         if self.unit.dimension != new_unit.dimension:
#             raise ValueError("Cannot change dimensions in conversion.")

#         # get the value in the base value
#         power = self.unit.unit_power
#         base_value = self.value * (self.unit.unit_prefix or 1)**power
#         factor = (self.unit.si_factor / new_unit.si_factor)
#         value = base_value * factor**power
#         return W24PhysicalQuantity(value=value, unit=new_unit)

#     def __str__(self) -> str:
#         unit_prefix = {
#             W24UnitPrefix.YOTTA: "Y",
#             W24UnitPrefix.ZETTA: "Z",
#             W24UnitPrefix.EXA: "E",
#             W24UnitPrefix.PETA: "P",
#             W24UnitPrefix.TERA: "T",
#             W24UnitPrefix.GIGA: "G",
#             W24UnitPrefix.MEGA: "M",
#             W24UnitPrefix.KILO: "k",
#             W24UnitPrefix.HECTO: "h",
#             W24UnitPrefix.DECA: "da",
#             None: "",
#             W24UnitPrefix.DECI: "d",
#             W24UnitPrefix.CENTI: "c",
#             W24UnitPrefix.MILLI: "m",
#             W24UnitPrefix.MICRO: "µ",
#             W24UnitPrefix.NANO: "n",
#             W24UnitPrefix.PICO: "p",
#             W24UnitPrefix.FEMTO: "f",
#             W24UnitPrefix.ATTO: "a",
#             W24UnitPrefix.ZEPTO: "z",
#             W24UnitPrefix.YOCTO: "y",
#         }.get(self.unit.unit_prefix)

#         # make sure we have actually found
#         # the value.
#         if unit_prefix is None:
#             raise ValueError("Unknown unit prefix")

#         return f"{self.value} {unit_prefix}{self.unit.symbol}"



from werk24.models.property.base import W24Property
from typing import Literal, Union, Optional
from pydantic import Field, BaseModel
from decimal import Decimal
from werk24.models.value import W24PhysicalQuantity, ureg


class W24Iso10110Grade(BaseModel):
    """ISO 10110 Grade.

    This model contains information about both, the
    homogeneity and the striae and is thus used by
    both models.
    """

    blurb: str = Field(examples=["2/3;1"])
    homogeneity_grade: Optional[str] = Field(examples=["3"], default=None)
    striae_grade: Optional[str] = Field(examples=["1"], default=None)


class W24Iso10110Limits(BaseModel):
    """ISO 10110 Limits for Homogeneity and Wavefront"""

    blurb: str = Field(examples=["5* 10^-6; <15nm"])
    tolerance_limit: Decimal = Field(examples=[Decimal(str("50e-6"))])
    striae_wavefront_deviation_tolerance_limit: Optional[W24PhysicalQuantity] = Field(
        examples=[W24PhysicalQuantity(blurb="15nm", value=15 * ureg.nm)],
        default=None
    )


class W24PropertyGlasHomogeneity(W24Property):
    # WORKAROUND FOR TYPO
    pass


class W24PropertyGlassHomogeneity(W24Property):
    """Parent for all Glass Homogeneity Properties"""

    property_type: Literal["GLASS_HOMOGENEITY"] = "GLASS_HOMOGENEITY"


class W24PropertyGlassHomogeneitySchottGrade(W24PropertyGlassHomogeneity):
    """Schott Homogeneity Grade"""

    property_subtype: Literal["SCHOTT_GRADE"] = "SCHOTT_GRADE"
    blurb: str = Field(examples=["H1", "H2", "H3"])
    grade: str = Field(examples=["H1", "H2", "H3"])


class W24PropertyGlassHomogeneityIso10110Grade(W24PropertyGlassHomogeneity):
    """ISO 10110 Homogeneity Grade"""

    property_subtype: Literal["ISO_10110_GRADE"] = "ISO_10110_GRADE"
    blurb: str = Field(examples=["2/3;1"])
    grade: W24Iso10110Grade


class W24PropertyGlassHomogeneityIso10110ToleranceLimit(W24PropertyGlassHomogeneity):
    """ISO 10110 Homogeneity Tolerance Limit"""

    property_subtype: Literal["ISO_10110_TOLERANCE_LIMIT"] = "ISO_10110_TOLERANCE_LIMIT"
    blurb: str = Field(examples=["5* 10^-6; <15nm"])
    limits: W24Iso10110Limits


class W24PropertyGlassHomogeneityIso10110NhGrade(W24PropertyGlassHomogeneity):
    """ISO 10110 Homogeneity NH Grade"""

    property_subtype: Literal["ISO_10110_NH_GRADE"] = "ISO_10110_NH_GRADE"
    blurb: str = Field(examples=["NH040"])
    grade: str = Field(examples=["NH040"])
    tolerance_limit: Decimal = Field(examples=[Decimal(str("40e-6"))])


class W24PropertyGlassHomogeneityFreeText(W24PropertyGlassHomogeneity):
    """Homogeneity Free Text"""

    free_text: str
    variation_type: Literal["FREETEXT"] = "FREETEXT"


W24PropertyGlassHomogeneityType = Union[
    W24PropertyGlassHomogeneitySchottGrade,
    W24PropertyGlassHomogeneityIso10110Grade,
    W24PropertyGlassHomogeneityIso10110ToleranceLimit,
    W24PropertyGlassHomogeneityIso10110NhGrade,
    W24PropertyGlassHomogeneityFreeText,
]



from enum import Enum


class W24AlignmentHorizontal(str, Enum):
    """Horizontal Alignment options"""

    LEFT = "LEFT"
    CENTER = "CENTER"
    RIGHT = "RIGHT"


class W24AlignmentVertical(str, Enum):
    """Vertical Alignment options"""

    TOP = "TOP"
    CENTER = "CENTER"
    BOTTOM = "BOTTOM"



""" Material Models
"""

from enum import Enum
from typing import Any, List, Optional, Tuple, Union, Literal
from pydantic import Field
from werk24.models.base_feature import BaseModel, W24BaseFeatureModel
from werk24.models.property.glass_homogeneity import W24PropertyGlassHomogeneityType
from werk24.models.typed_model import W24TypedModel


class W24MaterialCategory1(str, Enum):
    FERROUS_ALLOY = "FERROUS_ALLOY"
    NONFERROUS_ALLOY = "NONFERROUS_ALLOY"
    POLYMER = "POLYMER"
    CERAMIC = "CERAMIC"
    COMPOSITE = "COMPOSITE"
    ORGANIC = "ORGANIC"


class W24MaterialCategory2(str, Enum):
    # FERROUS ALLOYS
    STEEL = "STEEL"
    IRON = "IRON"
    MAGNETIC_OR_ELECTRICAL_MATERIAL = "MAGNETIC_OR_ELECTRICAL_MATERIAL"
    SINTERED_POWDERED_METAL = "SINTERED_POWDERED_METAL"
    WELDING_FILLER_MATERIAL = "WELDING_FILLER_MATERIAL"

    # NONFERROUS ALLOY
    ALUMINUM = "ALUMINUM"
    ANTIMONY = "ANTIMONY"
    BERYLLIUM = "BERYLLIUM"
    BISMUTH = "BISMUTH"
    CADMIUM = "CADMIUM"
    CHROMIUM = "CHROMIUM"
    COBALT = "COBALT"
    COPPER = "COPPER"
    GALLIUM = "GALLIUM"
    IRIDIUM = "IRIDIUM"
    LEAD = "LEAD"
    MAGNESIUM = "MAGNESIUM"
    MANGANESE = "MANGANESE"
    MERCURY = "MERCURY"
    MOLYBDENUM = "MOLYBDENUM"
    NICKEL = "NICKEL"
    NIOBIUM = "NIOBIUM"
    PLATINUM = "PLATINUM"
    REAR_EARTHS = "REAR_EARTHS"
    SELENIUM = "SELENIUM"
    SILICON = "SILICON"
    TANTALUM = "TANTALUM"
    THALLIUM = "THALLIUM"
    TIN = "TIN"
    TITANIUM = "TITANIUM"
    TUNGSTEN = "TUNGSTEN"
    VANADIUM = "VANADIUM"
    ZINC = "ZINC"
    ZIRCONIUM = "ZIRCONIUM"

    # POLYMER
    THERMOPLAST = "THERMOPLAST"
    THERMOSET = "THERMOSET"
    THERMOPLASTIC_ELASTOMER = "THERMOPLASTIC_ELASTOMER"
    THERMOSET_ELASTOMER = "THERMOSET_ELASTOMER"
    POLYMER_FOAM = "POLYMER_FOAM"

    # CERAMIC
    REFRACTORY = "REFRACTORY"
    TECHNICAL_CERAMIC = "TECHNICAL_CERAMIC"
    CERAMIC_FOAM = "CERAMIC_FOAM"

    # COMPOSITE
    MATRIX = "MATRIX"
    REINFORCEMENT = "REINFORCEMENT"

    # ORGANIC
    WOOD = "WOOD"


class W24MaterialCategory3(str, Enum):
    # FERROUS_ALLOY / STEEL
    STRUCTURAL_OR_CONSTRUCTIONAL_STEEL = "STRUCTURAL_OR_CONSTRUCTIONAL_STEEL"
    STAINLESS_STEEL = "STAINLESS_STEEL"
    TOOL_STEEL = "TOOL_STEEL"
    CAST_STEEL = "CAST_STEEL"

    # FERROUS_ALLOY / IRON
    CAST_IRON = "CAST_IRON"
    FERROALLOY = "FERROALLOY"

    # POLYMER / THERMOPLAST
    ABS = "ABS"
    ABS_PC = "ABS_PC"
    ABS_PLUS_PA = "ABS_PLUS_PA"
    ABS_PLUS_PA6 = "ABS_PLUS_PA6"
    ABS_PLUS_PA66 = "ABS_PLUS_PA66"
    ABS_PLUS_PBT = "ABS_PLUS_PBT"
    ABS_PLUS_PC = "ABS_PLUS_PC"
    ABS_PLUS_PET = "ABS_PLUS_PET"
    ABS_PLUS_PMMA = "ABS_PLUS_PMMA"
    ABS_PLUS_PVC = "ABS_PLUS_PVC"
    ABS_PLUS_PVC_C = "ABS_PLUS_PVC_C"
    ABS_PLUS_SAN = "ABS_PLUS_SAN"
    ABS_PLUS_TPE = "ABS_PLUS_TPE"
    ABS_PLUS_TPES = "ABS_PLUS_TPES"
    ABS_PLUS_TPU = "ABS_PLUS_TPU"
    ABS_PP = "ABS_PP"
    ACPES = "ACPES"
    AES = "AES"
    AES_PLUS_PC = "AES_PLUS_PC"
    ASA = "ASA"
    ASA_PC = "ASA_PC"
    ASA_PLUS_ABS = "ASA_PLUS_ABS"
    ASA_PLUS_AES = "ASA_PLUS_AES"
    ASA_PLUS_MSAN = "ASA_PLUS_MSAN"
    ASA_PLUS_PA = "ASA_PLUS_PA"
    ASA_PLUS_PC = "ASA_PLUS_PC"
    ASA_PLUS_PMMA = "ASA_PLUS_PMMA"
    ASA_PLUS_SAN = "ASA_PLUS_SAN"
    ASA_PLUS_TPE = "ASA_PLUS_TPE"
    BMI = "BMI"
    BVOH = "BVOH"
    CAB = "CAB"
    CAP = "CAP"
    CE = "CE"
    COC = "COC"
    CP = "CP"
    CPE = "CPE"
    CPT = "CPT"
    E_P = "E_P"
    EA = "EA"
    EAA = "EAA"
    EBA = "EBA"
    EBACO = "EBACO"
    EC = "EC"
    ECTFE = "ECTFE"
    EEA = "EEA"
    EFEP = "EFEP"
    EMAA = "EMAA"
    EMAAA = "EMAAA"
    ENBAGMA = "ENBAGMA"
    ETFE = "ETFE"
    EVA = "EVA"
    EVACO = "EVACO"
    EVOH = "EVOH"
    FEP = "FEP"
    HDPE = "HDPE"
    HIPP = "HIPP"
    HIPS = "HIPS"
    IONOMER_RESIN = "IONOMER_RESIN"
    LCP = "LCP"
    LCP_PLUS_PPS = "LCP_PLUS_PPS"
    LDPE = "LDPE"
    LDPE_EVA = "LDPE_EVA"
    LDPE_LLDPE = "LDPE_LLDPE"
    LLDPE = "LLDPE"
    LMDPE = "LMDPE"
    MABS = "MABS"
    MBS = "MBS"
    MDPE = "MDPE"
    MFA = "MFA"
    PA = "PA"
    PA_MACM10_1010 = "PA_MACM10_1010"
    PA_MACM12 = "PA_MACM12"
    PA_MACM12_PLUS_PA_12 = "PA_MACM12_PLUS_PA_12"
    PA_MXD6 = "PA_MXD6"
    PA_MXD6_MXDI = "PA_MXD6_MXDI"
    PA_PACM12 = "PA_PACM12"
    PA_PLUS_HDPE = "PA_PLUS_HDPE"
    PA_PLUS_PA66_6 = "PA_PLUS_PA66_6"
    PA_PLUS_PE = "PA_PLUS_PE"
    PA_PLUS_PP = "PA_PLUS_PP"
    PA_PLUS_PPA = "PA_PLUS_PPA"
    PA_PLUS_SPS = "PA_PLUS_SPS"
    PA_PLUS_TPE = "PA_PLUS_TPE"
    PA_TPE = "PA_TPE"
    PA1010 = "PA1010"
    PA1012 = "PA1012"
    PA10T_X = "PA10T_X"
    PA11 = "PA11"
    PA12 = "PA12"
    PA12_MACMI = "PA12_MACMI"
    PA12_MACMI_PLUS_PA12 = "PA12_MACMI_PLUS_PA12"
    PA410 = "PA410"
    PA46 = "PA46"
    PA46_PLUS_PA6 = "PA46_PLUS_PA6"
    PA4T = "PA4T"
    PA510 = "PA510"
    PA6 = "PA6"
    PA6_12 = "PA6_12"
    PA6_3T = "PA6_3T"
    PA6_66_12 = "PA6_66_12"
    PA6_66_136 = "PA6_66_136"
    PA6_66_610 = "PA6_66_610"
    PA6_69 = "PA6_69"
    PA6_6I = "PA6_6I"
    PA6_6T = "PA6_6T"
    PA6_IPDI = "PA6_IPDI"
    PA6_PA12 = "PA6_PA12"
    PA6_PLUS_ASA = "PA6_PLUS_ASA"
    PA6_PLUS_PA_6I_6T = "PA_6_PLUS_PA_6I_6T"
    PA6_PLUS_PA12 = "PA6_PLUS_PA12"
    PA6_PLUS_PA12_X = "PA6_PLUS_PA12_X"
    PA6_PLUS_PA66 = "PA6_PLUS_PA66"
    PA6_PLUS_PA66_6 = "PA6_PLUS_PA66_6"
    PA6_PLUS_PE = "PA6_PLUS_PE"
    PA6_PLUS_PP = "PA6_PLUS_PP"
    PA610 = "PA610"
    PA612 = "PA612"
    PA66 = "PA66"
    PA66_6 = "PA66_6"
    PA66_610 = "PA66_610"
    PA66_PLUS_PA610 = "PA66_PLUS_PA610"
    PA66_PLUS_PA612 = "PA66_PLUS_PA612"
    PA66_PLUS_PA6I = "PA66_PLUS_PA6I"
    PA66_PLUS_PA6I_6T = "PA66_PLUS_PA6I_6T"
    PA66_PLUS_PA6I_X = "PA66_PLUS_PA6I_X"
    PA66_PLUS_PE = "PA66_PLUS_PE"
    PA66_PLUS_PP = "PA66_PLUS_PP"
    PA66_PLUS_PPA = "PA66_PLUS_PPA"
    PA69 = "PA69"
    PA6I = "PA6I"
    PA6I_6T = "PA6I_6T"
    PA6I_X = "PA6I_X"
    PA6T = "PA6T"
    PA6T_66 = "PA6T_66"
    PA6T_XT_PLUS_PA6T_66 = "PA6T_XT_PLUS_PA6T_66"
    PA6T_6I = "PA6T_6I"
    PA6T_6I_66 = "PA6T_6I_66"
    PA6T_MPMDT = "PA6T_MPMDT"
    PA6T_XT = "PA6T_XT"
    PA9T = "PA9T"
    PAEK = "PAEK"
    PAI = "PAI"
    PAMXD6_PA66 = "PAMXD6_PA66"
    PAO = "PAO"
    PAR = "PAR"
    PARA = "PARA"
    PB = "PB"
    PBAT = "PBAT"
    PBAT_PLUS_PLA = "PBAT_PLUS_PLA"
    PBI = "PBI"
    PBS = "PBS"
    PBT = "PBT"
    PBT_ASA = "PBT_ASA"
    PBT_PC = "PBT_PC"
    PBT_PLUS_ASA = "PBT_PLUS_ASA"
    PBT_PLUS_PE = "PBT_PLUS_PE"
    PBT_PLUS_PET = "PBT_PLUS_PET"
    PBT_PLUS_PET_PLUS_ASA = "PBT_PLUS_PET_PLUS_ASA"
    PBT_PLUS_PETG = "PBT_PLUS_PETG"
    PBT_PLUS_PPE = "PBT_PLUS_PPE"
    PBT_PLUS_PS = "PBT_PLUS_PS"
    PBT_PLUS_SAN = "PBT_PLUS_SAN"
    PC = "PC"
    PC_PET = "PC_PET"
    PC_PLUS_HIPS = "PC_PLUS_HIPS"
    PC_PLUS_MBS = "PC_PLUS_MBS"
    PC_PLUS_PBT = "PC_PLUS_PBT"
    PC_PLUS_PCT = "PC_PLUS_PCT"
    PC_PLUS_PET = "PC_PLUS_PET"
    PC_PLUS_PET_PLUS_PBT = "PC_PLUS_PET_PLUS_PBT"
    PC_PLUS_PLA = "PC_PLUS_PLA"
    PC_PLUS_PMMA = "PC_PLUS_PMMA"
    PC_PLUS_PPC = "PC_PLUS_PPC"
    PC_PLUS_PS = "PC_PLUS_PS"
    PC_PLUS_PTFE = "PC_PLUS_PTFE"
    PC_PLUS_SAN = "PC_PLUS_SAN"
    PC_PLUS_TPES = "PC_PLUS_TPES"
    PC_PLUS_TPU = "PC_PLUS_TPU"
    PC_PPC = "PC_PPC"
    PCL = "PCL"
    PCT = "PCT"
    PCTA = "PCTA"
    PCTFE = "PCTFE"
    PCTG = "PCTG"
    PCTG_PLUS_PC = "PCTG_PLUS_PC"
    PE = "PE"
    PE_HMW = "PE_HMW"
    PE_HMWHD = "PE_HMWHD"
    PE_PLUS_HIPS = "PE_PLUS_HIPS"
    PE_UHMW = "PE_UHMW"
    PEEK = "PEEK"
    PEI = "PEI"
    PEI_PLUS_PCE = "PEI_PLUS_PCE"
    PEK = "PEK"
    PEKEKK = "PEKEKK"
    PEKK = "PEKK"
    PEN = "PEN"
    PESU = "PESU"
    PET = "PET"
    PET_G = "PET_G"
    PET_PLUS_PA6 = "PET_PLUS_PA6"
    PET_PLUS_PA66 = "PET_PLUS_PA66"
    PEX_B = "PEX_B"
    PFA = "PFA"
    PFSA_PTFE = "PFSA_PTFE"
    PGA = "PGA"
    PHA = "PHA"
    PHB = "PHB"
    PHBV = "PHBV"
    PI = "PI"
    PIB = "PIB"
    PK = "PK"
    PLA = "PLA"
    PLA_PCL = "PLA_PCL"
    PLA_PEG = "PLA_PEG"
    PLA_PHA = "PLA_PHA"
    PLA_PLUS_ABS = "PLA_PLUS_ABS"
    PLA_PLUS_HDPE = "PLA_PLUS_HDPE"
    PLA_PLUS_PHB = "PLA_PLUS_PHB"
    PLA_PLUS_PMMA = "PLA_PLUS_PMMA"
    PLGA = "PLGA"
    PMMA = "PMMA"
    PMMI = "PMMI"
    PMP = "PMP"
    PMS = "PMS"
    POM = "POM"
    POM_PLUS_MBS = "POM_PLUS_MBS"
    POM_PLUS_PE = "POM_PLUS_PE"
    POM_PLUS_PTFE = "POM_PLUS_PTFE"
    POM_PLUS_PUR = "POM_PLUS_PUR"
    PP = "PP"
    PP_PE = "PP_PE"
    PP_PLUS_EPDM = "PP_PLUS_EPDM"
    PP_PLUS_EPR = "PP_PLUS_EPR"
    PP_PLUS_EVA = "PP_PLUS_EVA"
    PP_PLUS_PE = "PP_PLUS_PE"
    PP_PLUS_PPE_PLUS_PS = "PP_PLUS_PPE_PLUS_PS"
    PPA = "PPA"
    PPC = "PPC"
    PPE = "PPE"
    PPE_PLUS_HIPS = "PPE_PLUS_HIPS"
    PPE_PLUS_PA = "PPE_PLUS_PA"
    PPE_PLUS_PA6 = "PPE_PLUS_PA6"
    PPE_PLUS_PA66 = "PPE_PLUS_PA66"
    PPE_PLUS_PP = "PPE_PLUS_PP"
    PPE_PLUS_PS = "PPE_PLUS_PS"
    PPE_PLUS_PS_PLUS_PA = "PPE_PLUS_PS_PLUS_PA"
    PPE_PLUS_TPE = "PPE_PLUS_TPE"
    PPE_PLUS_TPS_SEBS = "PPE_PLUS_TPS_SEBS"
    PPOX = "PPOX"
    PPS = "PPS"
    PPS_PLUS_PA = "PPS_PLUS_PA"
    PPS_PLUS_PPE = "PPS_PLUS_PPE"
    PPS_PLUS_PTFE = "PPS_PLUS_PTFE"
    PPSU = "PPSU"
    PPSU_PLUS_PSU = "PPSU_PLUS_PSU"
    PS = "PS"
    PS_I = "PS_I"
    PS_PLUS_PE = "PS_PLUS_PE"
    PS_PLUS_PMMA = "PS_PLUS_PMMA"
    PS_PLUS_SPS = "PS_PLUS_SPS"
    PSU = "PSU"
    PSU_PLUS_ABS = "PSU_PLUS_ABS"
    PSU_PLUS_PC = "PSU_PLUS_PC"
    PTFE = "PTFE"
    PTT = "PTT"
    PVAL = "PVAL"
    PVB = "PVB"
    PVC = "PVC"
    PVC_C = "PVC_C"
    PVC_C_PLUS_PVC = "PVC_C_PLUS_PVC"
    PVC_PLUS_NBR = "PVC_PLUS_NBR"
    PVC_PLUS_PMMA = "PVC_PLUS_PMMA"
    PVC_PLUS_PUR = "PVC_PLUS_PUR"
    PVC_U = "PVC_U"
    PVCA = "PVCA"
    PVDC = "PVDC"
    PVDF = "PVDF"
    PVP = "PVP"
    RPS = "RPS"
    SAN = "SAN"
    SB = "SB"
    SMA = "SMA"
    SMA_PLUS_HIPS = "SMA_PLUS_HIPS"
    SMI = "SMI"
    SMMA = "SMMA"
    SPS = "SPS"
    SRP = "SRP"
    TPC_EE_PLUS_PBT = "TPC_EE_PLUS_PBT"
    TPC_ET_PLUS_EMA = "TPC_ET_PLUS_EMA"
    TPC_ET_PLUS_PBT = "TPC_ET_PLUS_PBT"
    TPES = "TPES"
    TPU_PLUS_TPS_PLUS_TPA = "TPU_PLUS_TPS_PLUS_TPA"
    VDF_CTFE = "VDF_CTFE"
    VDF_HFP = "VDF_HFP"
    VINYL_CHLORIDE_BLEND = "VINYL_CHLORIDE_BLEND"
    VLDPE = "VLDPE"
    XLPE = "XLPE"
    XLPO = "XLPO"

    # POLYMER / THERMOSET
    CA = "CA"
    EP = "EP"
    MA = "MA"
    MF = "MF"
    MP = "MP"
    PDAIP = "PDAIP"
    PDAP = "PDAP"
    PUR = "PUR"
    UF = "UF"
    UP = "UP"
    VE = "VE"

    # POLYMER / THERMOPLASTIC_ELASTOMER
    MPR = "MPR"
    PCU = "PCU"
    PEBA = "PEBA"
    SPU = "SPU"
    TPA = "TPA"
    TPC = "TPC"
    TPC_EE = "TPC_EE"
    TPC_ES = "TPC_ES"
    TPC_ET = "TPC_ET"
    TPE = "TPE"
    TPO = "TPO"
    TPO_EB = "TPO_EB"
    TPO_EO = "TPO_EO"
    TPO_EPDM_PLUS_PP = "TPO_EPDM_PLUS_PP"
    TPS = "TPS"
    TPS_SBS = "TPS_SBS"
    TPS_SEBS = "TPS_SEBS"
    TPS_SIS = "TPS_SIS"
    TPU = "TPU"
    TPU_ALES = "TPU_ALES"
    TPU_ALET = "TPU_ALET"
    TPU_ARES = "TPU_ARES"
    TPU_ARET = "TPU_ARET"
    TPU_ES = "TPU_ES"
    TPU_ET = "TPU_ET"
    TPV = "TPV"
    TPV_EPDM = "TPV_EPDM"
    TPV_EPDM_PLUS_PP = "TPV_EPDM_PLUS_PP"
    TPV_IIR_HIIR = "TPV_IIR_HIIR"
    TPV_NBR = "TPV_NBR"
    TSPCU = "TSPCU"
    TSPU = "TSPU"

    # POLYMER / THERMOSET_ELASTOMER
    ACM = "ACM"
    ACSM = "ACSM"
    AEM = "AEM"
    AU = "AU"
    BIIR = "BIIR"
    BR = "BR"
    CIIR = "CIIR"
    CM = "CM"
    CR = "CR"
    CR_NBR = "CR_NBR"
    CR_SBR = "CR_SBR"
    CSM = "CSM"
    ECO = "ECO"
    EPDM = "EPDM"
    EPDM_SBR = "EPDM_SBR"
    EPM = "EPM"
    EPM_EPDM = "EPM_EPDM"
    EPT = "EPT"
    EU = "EU"
    FFKM = "FFKM"
    FKM = "FKM"
    FMQ = "FMQ"
    FVMQ = "FVMQ"
    HNBR = "HNBR"
    IIR = "IIR"
    IR = "IR"
    IR_BR = "IR_BR"
    NBR = "NBR"
    NBR_PVC = "NBR_PVC"
    NBR_SBR = "NBR_SBR"
    NR = "NR"
    PDMS = "PDMS"
    PUMA = "PUMA"
    PVMQ = "PVMQ"
    RET = "RET"
    SBR = "SBR"
    SBR_ACM = "SBR_ACM"
    SBR_BR = "SBR_BR"
    SBR_IR = "SBR_IR"
    SBR_NR = "SBR_NR"
    SI = "SI"
    VMQ = "VMQ"
    VMQ_PVMQ = "VMQ_PVMQ"
    XNBR = "XNBR"
    XNBR_PVC = "XNBR_PVC"

    # COMPOSITE
    CARBON_FIBER_COMPOSITE = "CARBON_FIBER_COMPOSITE"
    GLASS_FIBER_COMPOSITE = "GLASS_FIBER_COMPOSITE"

    # REINFORCEMENT
    FIBER_REINFORCEMENT = "FIBER_REINFORCEMENT"
    STRUCTURAL_COMPOSITE = "STRUCTURAL_COMPOSITE"

    # FOAM / POLYMER_FOAM
    POLYPHENYLENE_OR_POLYSTYRENE = "POLYPHENYLENE_OR_POLYSTYRENE"
    POLYPHENYLENE_OR_POLYSTYRENE_HIPS = "POLYPHENYLENE_OR_POLYSTYRENE_HIPS"
    POLYCARBONATE = "POLYCARBONATE"

    # CERAMIC / REFRACTORY
    ALUMINA_BASED_CERAMIC = "ALUMINA_BASED_CERAMIC"
    CALCIA_BASED_CERAMIC = "CALCIA_BASED_CERAMIC"
    CARBON_BASED_CERAMIC = "CARBON_BASED_CERAMIC"
    CHROMIA_BASED_CERAMIC = "CHROMIA_BASED_CERAMIC"
    MAGNESIA_BASED_CERAMIC = "MAGNESIA_BASED_CERAMIC"
    MONOLOTHIC = "MONOLOTHIC"
    SIC_BASED_CERAMIC = "SIC_BASED_CERAMIC"
    SILICA_BASED_CERAMIC = "SILICA_BASED_CERAMIC"
    ZIRCONIA_BASED_CERAMIC = "ZIRCONIA_BASED_CERAMIC"

    # CERAMIC / TECHNICAL_CERAMIC
    BORIDE_CERAMIC = "BORIDE_CERAMIC"
    CARBIDE_CERAMIC = "CARBIDE_CERAMIC"
    GLASS_CERAMIC = "GLASS_CERAMIC"
    MIXED_CERAMIC = "MIXED_CERAMIC"
    NITRIDE_CERAMIC = "NITRIDE_CERAMIC"
    SILICATE_CERAMIC = "SILICATE_CERAMIC"
    SINGLE_OXIDE_CERAMIC = "SINGLE_OXIDE_CERAMIC"
    TITANATE_CERAMIC = "TITANATE_CERAMIC"

    # CERAMIC / FIBER
    FIBER_CERAMIC = "FIBER_CERAMIC"


class W24MaterialConditionBase(W24TypedModel):
    class Config:
        discriminators: Tuple[str, ...] = ("condition_type",)

    condition_type: str
    blurb: str


class W24AluminumTemper(W24MaterialConditionBase):
    """
    Aluminium Temper

    Following ISO 2107
    """

    condition_type: Literal["ALUMINUM_TEMPER"] = "ALUMINUM_TEMPER"

    blurb: str = Field(
        description="Blurb of the Aluminum Temper",
        examples=["T6", "H32"],
    )


class W24SteelTreatment(W24MaterialConditionBase):
    """
    Steel Treatments

    Following EN 10083
    """

    condition_type: Literal["STEEL_TREATMENT"] = "STEEL_TREATMENT"
    blurb: str = Field(
        description="Blurb of Steel Treatments",
        examples=["+Q"],
    )


class W24GlassHomogeneityCondition(W24MaterialConditionBase):
    """
    Glass Homogeneity Condition
    """

    condition_type: Literal["GLASS_HOMOGENEITY"] = "GLASS_HOMOGENEITY"
    glass_homogeneity: W24PropertyGlassHomogeneityType


""" List of the Material Conditions for the different material types """
W24MaterialCondition = Union[
    W24GlassHomogeneityCondition,
    W24AluminumTemper,
    W24SteelTreatment,
]


class W24Material(W24BaseFeatureModel):
    """W24 Object for Materials.

    Parsed Material object that can either be
    associated to the TitleBlock or derived from
    all the available information (including the
    text on the canvas.

    Attributes:
    ----------
    blurb: Material Name for human consumption.
        This will typically include the designation
        and the standard.

    raw_ocr_blurb: Material Name as it was indicated
        on the drawing. This contains more information
        than the blurb and is also present when the
        material was not found in the internal
        material database.

    standard: Material Standard indicated
        on the technical drawing. This used to be
        an enum - but we now have over 100 supported
        material standards and the number is increasing
        weekly. So a string seems to be more appropriate.

    designation: Name of the material in accordance
        with the material standard. An alternative name
        would be material_designation.

    material_category: Categorization of the material
        following W24MaterialCategoryX Tree.

    material_conditions: list of the material conditions.
        The conditions are specific to the material category
        and can be used to further specify the material
        properties.
    """

    blurb: str
    raw_ocr_blurb: str = ""
    standard: str
    designation: str
    material_category: Tuple[
        Optional[W24MaterialCategory1],
        Optional[W24MaterialCategory2],
        Optional[W24MaterialCategory3],
    ]
    material_conditions: list[W24MaterialCondition] = []

    # Deprecated since 1.4.0. Set to default=None to avoid breaking changes
    material_standard: Optional[Any] = None
    material_code: Optional[Any] = ""


class W24MaterialSet(BaseModel):
    """Set of Materials are used when two or more materials are defined
        for a part which are applicable together.

    Args:
    ----
        material (list[W24material]): List of W24Materials that are
            defined together for a part.
            For example,
                Material_A+Material_B+MAterial_C
                Commonly occurs with polymers like PA+PVC+GF

    """

    material: List[W24Material]
    blurb: str
